/* -*- Mode: C++ -*- */

/* geometry.h
 * CMUnited99 (soccer client for Robocup99)
 * Peter Stone <pstone@cs.cmu.edu>
 * Computer Science Department
 * Carnegie Mellon University
 * Copyright (C) 1999 Peter Stone
 *
 * CMUnited-99 was created by Peter Stone, Patrick Riley, and Manuela Veloso
 *
 * You may copy and distribute this program freely as long as you retain this notice.
 * If you make any changes or have any comments we would appreciate a message.
 * For more information, please see http://www.cs.cmu.edu/~robosoccer/
 */


#ifndef _GEOMETRY_
#define _GEOMETRY_

#include "utils.h"

/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/*                        Vector Class                                          */
/********************************************************************************/
/********************************************************************************/
/********************************************************************************/


class CMUVector {
public:
  float x;
  float y;
  
  CMUVector(float vx = 0, float vy = 0) { x = vx; y = vy; }
  CMUVector operator -() { return CMUVector(-x, -y); }
  CMUVector operator +(const CMUVector &a) { return CMUVector(x + a.x, y + a.y); }
  CMUVector operator -(const CMUVector &a) { return CMUVector(x - a.x, y - a.y); }
  CMUVector operator *(const float &a) { return CMUVector(x * a, y * a); }
  CMUVector operator *(const CMUVector &a) { return CMUVector(x * a.x, y * a.y); }
  CMUVector operator /(const float &a) { return CMUVector(x / a, y / a); }
  CMUVector operator /(const CMUVector &a) { return CMUVector(x / a.x, y / a.y); }
  void operator =(const float &a) { x = a; y = a; }
  void operator +=(const CMUVector &a) { x += a.x;  y += a.y; }
  void operator +=(const float &a) { x += a;  y += a; }
  void operator -=(const CMUVector &a) { x -= a.x;  y -= a.y; }
  void operator -=(const float &a) { x -= a;  y -= a; }
  void operator *=(const float &a) { x *= a;  y *= a; }
  void operator /=(const float &a) { x /= a;  y /= a; }
  bool operator !=(const CMUVector &a) { return (x != a.x) || (y != a.y); } 
  bool operator !=(const float &a) { return (x != a) || (y != a); }
  bool operator ==(const CMUVector &a) { return (x == a.x) && (y == a.y); } 
  friend ostream& operator <<(ostream & os, CMUVector v) { return os << "("<<v.x<<", "<<v.y<<")"; } 
  void Print() { cout << *this << endl; }
  void ins(float vx = 0, float vy = 0) { x = vx; y = vy; }
  float mod() { return sqrt(x*x + y*y); }
  float mod2() { return x*x + y*y; }
  float dist(const CMUVector &a) { return (*this-a).mod(); }
  float dist2(const CMUVector &a) { return (*this-a).mod2(); }
  CMUVector Normalize() { return SetLength(1.0); }
  CMUVector SetLength(float len) { return (*this) * (len / mod()); }
  AngleDeg dir() { return ATan2(y,x); }
  CMUVector rotate(AngleDeg angle) 
  { return CMUVector(this->mod()*Cos(this->dir()+angle), this->mod()*Sin(this->dir()+angle)); }
  CMUVector Global2Relative(CMUVector orig, float ang);
  CMUVector Relative2Global(CMUVector orig, float ang);
  Bool InFrontOf(const CMUVector &a) { return x > a.x ? TRUE : FALSE; }
  Bool InFrontOf(const float &a) { return x > a ? TRUE : FALSE; }
  Bool Behind(const CMUVector &a) { return x < a.x ? TRUE : FALSE; }
  Bool Behind(const float &a) { return x < a ? TRUE : FALSE; }

  Bool ApproxEqual(const CMUVector &a) 
    { return (fabs(x-a.x) < FLOAT_EPS && fabs(y-a.y) < FLOAT_EPS) ? TRUE : FALSE; }

  CMUVector WeightedAverage(const CMUVector &a, float w)
    { return CMUVector((1.0-w)*x + w * a.x, (1.0-w)*y + w * a.y); }
};

inline CMUVector Polar2CMUVector(float mod, AngleDeg ang) {
  return CMUVector(mod * Cos(ang), mod * Sin(ang));
}
  
inline CMUVector CMUVector::Global2Relative(CMUVector orig, float ang)
{
  return (*this - orig).rotate(-ang);
}

inline CMUVector CMUVector::Relative2Global(CMUVector orig, float ang)
{
  return (rotate(ang) + orig);
}

inline CMUVector operator *(const float &a, const CMUVector &v)
{ return CMUVector(v.x * a, v.y * a); }


/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/*                       Ray Class                                              */
/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/* stores origin and direction (which is normalized) */
class Line;
class Rectangle;
class Ray 
{
public:
  Ray() 
    { origin = CMUVector(0,0); direction = CMUVector(1,0); }
  Ray(CMUVector orig, CMUVector dir);
  Ray(CMUVector orig, float ang)
    { origin = orig; direction = Polar2CMUVector(1.0, ang); }

  Bool OnRay(CMUVector pt);
  Bool InRightDir(CMUVector pt); // more lenient than above about distance off ray 
  
  Bool intersection(Ray r, CMUVector *pPt);
  Bool intersection(Line l, CMUVector *pPt);

  int CircleIntersect(float rad, CMUVector center,
		      CMUVector* psol1, CMUVector* psol2);

  CMUVector GetClosestPoint(CMUVector pt);

  CMUVector RectangleIntersection(Rectangle R);
  
protected:
  friend Line;
  friend Rectangle;
  CMUVector origin;
  CMUVector direction;
};

/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/*                      Line Class                                              */
/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
class Line {
 public:
  Line(float x_coef=0.0, float y_coef=0.0, float constant=0.0);
  Line(Ray r) 
    { LineFromRay(r); }
  
  void LineFromTwoPoints(CMUVector pt1, CMUVector pt2);
  void LineFromRay(Ray r);
  void LineFromRay(CMUVector orig, CMUVector dir)
    { LineFromRay(Ray(orig, dir)); }
  void LineFromRay(CMUVector orig, AngleDeg dir)
    { LineFromRay(orig, Polar2CMUVector(1.0, dir)); }
  
  Bool PointOnLine(float x, float y);
  inline Bool PointOnLine(CMUVector pt) { return PointOnLine(pt.x, pt.y); }

  float dist(CMUVector pt);
  float dist2(CMUVector pt);
  float angle(); /* returns the angle of the line in [-90, 90] */
  
  Bool InBetween(CMUVector pt, CMUVector end1, CMUVector end2);
  CMUVector GetClosestPtInBetween(CMUVector pt, CMUVector end1, CMUVector end2);
  
  /* the buffer should really be linked to an option or something,
     but 1.0 is okay */
  inline Bool OnLine(CMUVector pt, float buffer = 1.0)
    { return (dist2(pt) < Sqr(buffer)) ? TRUE : FALSE; }
  
  CMUVector ProjectPointUsingCircle(CMUVector pt);
  inline CMUVector ProjectPoint(CMUVector pt) { return intersection(perpendicular(pt)); }

  float get_y(float x);
  float get_x(float y);
  CMUVector intersection(Line l);
  inline Line perpendicular(CMUVector pt) { return Line(B,-A,A*pt.y - B*pt.x); }

  Bool RayIntersection(Ray r, CMUVector *ppt);
  
  Line shift_y(float val)
    { return Line(A, B, C - val * B); }
  Line shift_x(float val)
    { return Line(A, B, C - val * A); }  

  /* returns whether the projection of pt1 is closer to targ_pt than the
     projection of pt2 */
  Bool IsPtCloserToPtOnLine(CMUVector pt1, CMUVector pt2, CMUVector targ_pt);
  
  Bool HalfPlaneTest(CMUVector pt); //return TRUE on top/left part of plane
  Bool SameSlope(Line l);
  
  friend ostream& operator <<(ostream & os, Line l)
    { return os << "#L("<<l.A<<", "<<l.B<<", "<<l.C<<")"; } 
  void Print(void) { cout << *this << endl; }

  // private:
  float A,B,C; /* the three coeffs in the line equation */
  /* Ax + By + C = 0 */
} ;

inline Line LineFromTwoPoints(CMUVector pt1, CMUVector pt2)
{ Line l; l.LineFromTwoPoints(pt1,pt2); return l; }
inline Line LineFromRay      (Ray r)
{ Line l; l.LineFromRay     (r); return l; }
inline Line LineFromRay      (CMUVector orig, CMUVector dir)
{ Line l; l.LineFromRay     (orig,dir); return l; }
inline Line LineFromRay      (CMUVector orig, AngleDeg dir)
{ Line l; l.LineFromRay     (orig,dir); return l; }



/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/*                        Rectangle Class                                       */
/********************************************************************************/
/********************************************************************************/
/********************************************************************************/

class Rectangle {
public:

  Rectangle() ;
  Rectangle(const float left, const float right, 
	    const float top, const float bottom) ;
  Rectangle(const CMUVector center, const CMUVector size) ;

  bool operator ==(const Rectangle &a) 
  { return (left_x == a.left_x) && (right_x == a.right_x) && 
      (top_y == a.top_y) && (bottom_y == a.bottom_y); } 

  inline float TopY()     { return top_y; }
  inline float BottomY()  { return bottom_y; }
  inline float RightX()   { return right_x; }
  inline float LeftX()    { return left_x; }

  inline CMUVector TopLeftCorner() { return CMUVector(left_x, top_y); }
  inline CMUVector TopRightCorner() { return CMUVector(right_x, top_y); }
  inline CMUVector BottomLeftCorner() { return CMUVector(left_x, bottom_y); }
  inline CMUVector BottomRightCorner() { return CMUVector(right_x, bottom_y); }        

  inline Line LeftEdge() {return LineFromTwoPoints(TopLeftCorner(), BottomLeftCorner());}
  inline Line RightEdge() {return LineFromTwoPoints(TopRightCorner(), BottomRightCorner());}
  inline Line TopEdge() {return LineFromTwoPoints(TopLeftCorner(), TopRightCorner());}
  inline Line BottomEdge() {return LineFromTwoPoints(BottomLeftCorner(), BottomRightCorner());}

  Line GetEdge(int n);  //order: top, right, bot, left; starting at 0
  CMUVector GetPoint(int n); //order: TL, TR, BR, BL; starting at 0
  
  inline float Width()    { return right_x - left_x; }   
  inline float Height()   { return bottom_y - top_y; }  

  inline CMUVector Center()  { return CMUVector( LeftX() + Width()/2, TopY() + Height()/2 ); }

  Bool   IsWithin(const CMUVector& p) ;
  CMUVector nearestHEdge(const CMUVector& p) ;
  CMUVector nearestVEdge(const CMUVector& p) ;
  CMUVector nearestEdge(const CMUVector& p) ;

  float  DistanceToEdge(const CMUVector& p);
  CMUVector AdjustToWithin(const CMUVector& p);

  Line nearestHEdgeLine(const CMUVector& p) ;
  Line nearestVEdgeLine(const CMUVector& p) ;
  Line nearestEdgeLine(const CMUVector& p) ;
  
  CMUVector random() ;

  Rectangle shrink(float val)
    { return expand(-val); }
  Rectangle expand(float val);

  Rectangle expandLeft(float val)
    {return Rectangle(left_x - val, right_x, top_y, bottom_y);}
  Rectangle expandRight(float val)
    {return Rectangle(left_x, right_x + val, top_y, bottom_y);}
  Rectangle expandTop(float val)
    {return Rectangle(left_x, right_x, top_y - val, bottom_y);}
  Rectangle expandBottom(float val)
    {return Rectangle(left_x, right_x, top_y, bottom_y + val);}
  Rectangle shrinkLeft(float val)
    { return expandLeft(-val); }
  Rectangle shrinkRight(float val)
    { return expandRight(-val); }
  Rectangle shrinkTop(float val)
    { return expandTop(-val); }
  Rectangle shrinkBottom(float val)
    { return expandBottom(-val); }
  
  friend ostream& operator <<(ostream & os, Rectangle r)
    { return os << "RECTANGLE:  x = " << r.LeftX() << " to " << r.RightX()
		<< "   y = " << r.TopY() << " to " << r.BottomY(); }
  void Print();

  CMUVector RayIntersection(Ray r);
  
private:
  float left_x ;
  float right_x ;
  float top_y ;
  float bottom_y ;

} ;


/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/*                  Miscellaneous Geometry Functions                            */
/********************************************************************************/
/********************************************************************************/
/********************************************************************************/

int QuadraticFormula(float a, float b, float c, float* psol1, float* psol2);

inline int RayCircleIntersect(Ray r, float rad, CMUVector center,
			      CMUVector* psol1, CMUVector* psol2)
{ return r.CircleIntersect(rad, center, psol1, psol2); }
  

int LineCircleIntersect(Line l, float rad, CMUVector center,
			CMUVector* psol1, CMUVector* psol2);

CMUVector AdjustPtToRectOnLine(CMUVector pt, Rectangle r, Line l);

Bool InBetween(CMUVector pt, CMUVector end1, CMUVector end2);

CMUVector PointInBetween(CMUVector pt1, CMUVector pt2, float pt1dist=0);

AngleDeg AngleBisect(AngleDeg a1, AngleDeg a2);

CMUVector GetClosestPtInBetween(CMUVector pt, CMUVector end1, CMUVector end2);

Bool IsPointInCone(CMUVector pt, float wid_dist_ratio, CMUVector end, CMUVector vert);

#endif
