#ifndef _CMD_H_
#define _CMD_H_

#include "globaldef.h"
#include <iostream> 

/** 
Die Klasse Cmd_Base dient als Basisklasse fuer die Klassen:
  
Cmd_main, Cmd_Neck, Cmd_View, Cmd_Say

Cmd_Base implementiert einen Standardmechanismus fuer diese
Kommando-Klassen. Dieser besteht darin, dass man ein Kommando sperren und
setzen kann. 

Fuer den Anwender ist die Klasse Cmd von Bedeutung. Sie enthaelt im wesentlichen 
vier Felder, die den entsprechnden Kommandoklassen entsprechen.

Struktur:

                      Cmd_Base
               
                         |
      __________________/ \_________________
     |             |           |            |

  Cmd_Main    Cmd_Neck    Cmd_View     Cmd_Say


und 

                        Cmd
 
              enthaelt je ein Objekt vom Typ:
              
              Cmd_Main    
              Cmd_Neck    
              Cmd_View     
              Cmd_Say

 
Autor: ART!
*/

struct Cmd_Base {
  ///Initialzustand ist immer : (nicht gesperrt) und (Kommando nicht gesetzt)
  Cmd_Base() { unset_lock(); unset_cmd(); }
  ///Pruefe ob Objekt gesperrt ist
  bool is_lock_set() const { return lock_set; }
  ///Pruefe ob ein Kommando gesetzt ist, nur bei Ergebnis true ist in den
  ///abgeleitete Klassen das Auslesen der Werte sinnvoll.
  bool is_cmd_set()  const { return cmd_set; }
  /// setze Sperre, muss meinstens nicht explizit aufgerufen werden, da das
  /// Setzen von Kommandos in den abgeleiteten Klassen automatisch eine Sperre erzeugt
  void set_lock()   { lock_set= true; }
  /// entferne Sperre, wird evtl. nicht gebraucht
  void unset_lock() { lock_set= false; }
  /// Falls ein gesetztes Kommando doch nicht beruecksichtigt werden soll, kann 
  /// explizit mit dieser Methode tun
  void unset_cmd()  { check_lock(); cmd_set= false; } 
protected:
  /// Interner Schalter: Sperre (ein/aus)
  bool lock_set;
  /// Interner Schalter: Kommando (gesetzt/nicht gesetzt)
  bool cmd_set;
  /// Fehlermeldung falls auf ein gesperrtes Objekt geschrieben wird
  void check_lock() const
    { if (lock_set) cerr << "\n this command is locked"; }
  /// Fehlermeldung falls kein Kommando gesetzt wurde
  void check_cmd() const
    { if (!cmd_set) cerr << "\n not command was set"; }
};

/** 

Kommandos die vom Spieler im Spiel abgesetzt werden koennen. Es darf jeweils
nur ein Kommando gesetzt sein. Nach dem Setzen eines Kommandos wird automatisch 
eine Sperre gesetzt.

Winkel und Koordinaten werden entsprechend den Konventionen aus

mdpstate.h   <- siehe Doku. in dieser Datei

behandelt.
*/
struct Cmd_Main: public Cmd_Base{
  Cmd_Main(): Cmd_Base() {}
  ///Typen von Kommandos
  const static int TYPE_MOVETO   = 1;
  const static int TYPE_TURN 	 = 2;
  const static int TYPE_DASH 	 = 3;
  const static int TYPE_KICK 	 = 4;
  const static int TYPE_CATCH    = 5;    
  int get_type() const { return type; }

  void set_moveto(Value x, Value y) 
    { check_lock(); cmd_set= true; lock_set= true; type = TYPE_MOVETO; par_1=x;par_2= y; }
  void set_turn(Angle angle) 
    { check_lock(); cmd_set= true; lock_set= true; type = TYPE_TURN; par_ang= angle; }
  void set_dash(Value power) 
    { check_lock(); cmd_set= true; lock_set= true; type = TYPE_DASH; par_1= power; }
  void set_kick(Value power, Angle angle) 
    { check_lock(); cmd_set= true; lock_set= true; type = TYPE_KICK; par_1= power; par_ang= angle; }
  void set_catch(Angle angle) 
    { check_lock(); cmd_set= true; lock_set= true; type = TYPE_CATCH; par_ang= angle; }

  void get_moveto(Value & x, Value & y) const
    { check_type(TYPE_MOVETO); x= par_1; y= par_2; }
  void get_turn(Angle & angle) const 
    { check_type(TYPE_TURN); angle= par_ang; }
  void get_dash(Value & power) const 
    { check_type(TYPE_DASH); power= par_1; }
  void get_kick(Value & power, Angle & angle) const 
    { check_type(TYPE_KICK); power= par_1; angle= par_ang; }
  void get_catch(Angle & angle) const 
    { check_type(TYPE_CATCH); angle= par_ang; }

  friend ostream& operator<< (ostream& o, const Cmd_Main & cmd);
private:
  int type;
  Value par_1,par_2;
  Angle par_ang;
  void check_type (int t) const {
    check_cmd();
    if (type != t) cerr << "wrong type"; 
  }
};

struct Cmd_Neck: public Cmd_Base {
  Cmd_Neck(): Cmd_Base() {}
  void set_turn(Angle angle) 
    { check_lock(); cmd_set= true; lock_set= true; par_ang= angle; }
  void get_turn(Angle & angle) const
    { check_cmd();  angle= par_ang; }

  friend ostream& operator<< (ostream& o, const Cmd_Neck & cmd);
private:
  Angle par_ang;
};

struct Cmd_View: public Cmd_Base {
  Cmd_View(): Cmd_Base() {}
  const static int VIEW_ANGLE_WIDE     = WIDE;  //WIDE etc. define in globaldef.h
  const static int VIEW_ANGLE_NORMAL   = NORMAL;
  const static int VIEW_ANGLE_NARROW   = NARROW;
  const static int VIEW_QUALITY_HIGH   = HIGH;
  const static int VIEW_QUALITY_LOW    = LOW;
  void set_angle_and_quality(int ang, int quality) 
    { check_lock(); cmd_set= true; lock_set= true; view_angle= ang; view_quality= quality; }
  void get_angle_and_quality(int & ang, int & quality) const
    { check_cmd(); ang= view_angle; quality= view_quality; }

  friend ostream& operator<< (ostream& o, const Cmd_View & cmd);
private:
  int view_angle;
  int view_quality;
};

struct Cmd_Say: public Cmd_Base {
  Cmd_Say(): Cmd_Base() {}
  ~Cmd_Say() {}
  const static int MAXLEN= 0;
  void set_message(const char *str); 
  void set_message(const char *str, const int p_priority);
  int get_priority()  const {return priority;};
  const char* get_message() const { 
    return message;
  }
  
  friend ostream& operator<< (ostream& o, const Cmd_Say & cmd);
private:
  char message[MAXLEN+1];
  int priority;
};


/** Die Klasse Cmd ist die Standardschnittstelle zum Weltmodell
    bzw. Laufzeitumgebung des Agenten. 
    Alles was ein Agent in einem Zeitschritt bestimmen kann, wird in dieser
    Klasse festgehalten.
    Sie ist das Gegestueck zur Klasse MDPstate

                                       |
                  Weltmodell       |MDPstate>
                                       |       Taktikebene
               Laufzeitumgebung        |
                                     <Cmd|
                                       |
				       
				       
   Alle Konventionen bezueglich Koordinaten und Winkelangaben, die in der
   Datei   
   
   mdpstate.h 

   festgehalten sind, gelten damit auch fuer die Klasse Cmd.
*/
struct Cmd {
  Cmd_Main   cmd_main;
  Cmd_Neck   cmd_neck;
  Cmd_View   cmd_view;
  Cmd_Say    cmd_say;
  friend ostream& operator<< (ostream& o, const Cmd & cmd);
};

#endif
