#ifndef _MDP_INFO_
#define _MDP_INFO_

#include "globaldef.h"
#include "mdpstate.h"
#include "mdp_memory.h"
#include "Vector.h"
#include "base_moves.h"
#include "options.h"

//#include "types.h"


/** @ralf:
    This class allow access to the values of mdp-state all over the program, 
    because all variables and functions are static. You have only call 
    update(mdp-state) normaly at begin of  behave(...) to have actuall data.

    Some additional functions like "is_ball_kickable()" will help you to 
    program easier. 
*/

#define DECISION_TYPE_NONE 0
#define DECISION_TYPE_SCORE 1
#define DECISION_TYPE_PASS 2
#define DECISION_TYPE_DRIBBLE 3
#define DECISION_TYPE_STOPBALL 4
#define DECISION_TYPE_FACE_BALL 5
#define DECISION_TYPE_EXPECT_PASS 6
#define DECISION_TYPE_LEAVE_OFFSIDE 7
#define DECISION_TYPE_RUNTO 8
#define DECISION_TYPE_INTERCEPTBALL 9
#define DECISION_TYPE_INTERCEPTSLOWBALL 10
#define DECISION_TYPE_CLEARANCE 11
#define DECISION_TYPE_OFFSIDETRAP 12
#define DECISION_TYPE_DEFEND 13
#define DECISION_TYPE_BLOCKBALLHOLDER 14
#define DECISION_TYPE_ATTACKBALLHOLDER 15
#define DECISION_TYPE_WAITANDSEE 16
#define DECISION_TYPE_SEARCH_BALL 17
#define DECISION_TYPE_DEFEND_FACE_BALL 18
#define DECISION_TYPE_ATTACK 19
#define DECISION_TYPE_OFFER_FOR_PASS 20
#define DECISION_TYPE_GOALIECLEARANCE 21
#define DECISION_TYPE_LOOKFORGOALIE 22
#define DECISION_TYPE_INTROUBLE 23

typedef struct{
  int type;
  int time;
  float p1,p2,p3,p4,p5;
} IntentionType;



class mdpInfo {

 public:
  /** constant meaning that any number of  steps can be done until stamina is below recovery_threshold
      occurs for example near the end of a halftime.
   */
  static const int stamina4infinite_steps;
   /** To have access of all values of the MDPstate
   */
  static MDPstate* mdp;

  /** information about current decision
   */
  //static IntentionType decision_info;
  static IntentionType intention[11];

  static Cmd *my_current_cmd;

  /** To have access of all values of the server state - special feature
   */
  static MDPstate* server_state;

  /** To have access of all values of the MDPmemory
   */
  static MDPmemory* memory;
  /** To set the actual MDPstate, which be used by all functions and in the hole agent.
   */
  static void update( MDPstate &m);
  /** To init mdpInfo.
   */
  static void init();
  /** Clear intentions of mine 
   */
  static void clear_my_intention();
  /** Clear intentions of my team
   */
  static void clear_intentions();
  /** set intention of teammate player_no */
  static void set_intention(int player_no, int type, int time, float p1=0., float p2=0., float p3=0., 
			    float p4=0., float p5 = 0.);
  /** set intention of teammate player_no by parsing a heart information */
  static void set_intention_by_heart_info(char* heartinfo);
  /** returns intention type of teammate player_no and also returns its intention 
   if player == me, then my current decision is returned, otherwise heart information from teammates
   is given */
  static int get_intention(int player_no, IntentionType &player_intention);
  /** returns intention type of teammate player_no 
   if player == me, then my current decision is returned, otherwise heart information from teammates
   is given */
  static int get_intention(int player_no);
  /** returns my intention type also returns intention */ 
  static int get_my_intention(IntentionType &player_intention);
  /** returns my intention type */
  static int get_my_intention();
  /** set my intention */
  static void set_my_intention(int type, float p1=0., float p2=0., float p3=0., 
			       float p4=0., float p5 = 0.);


  /** To set the actual MDPmemory, which be used by all functions and in the hole agent.
   */
  static void update_memory( MDPmemory &m);

  // ***********************************************************************************
  // **   Now follows additional functions
  // **
  // **   pos = position
  // **   vel = velosity
  // **   abs = absolute
  // **   rel = relative to the player itself (depends on own angle and positon)

  // 1. Functions for the player itself:
  /** checks whether my position has been accurately determined. */
  static Bool is_my_pos_valid();
  /** gets my absolute position. */
  static Vector my_pos_abs();
  /** checks whether I am the goalie. */
  inline static Bool am_I_goalie(){return (ClientOptions::consider_goalie == 1);};
  /** checks whether I am in an offside position. */
  static Bool am_I_offside();
  /** checks whether my velocity has been accurately determined. */
  static Bool is_my_vel_valid();
  /** gets my absolute velocity vector. */
  static Vector my_vel_abs();
  static Vector my_expected_pos_abs_with_dash(Value dash_power);
  static Vector my_expected_vel_with_dash(Value dash_power);
  /** gets width of view angle  of the player as a radian
      wide -> PI
      normal -> PI/2.0
      narrow -> PI/4.0
      ServerOptions::visible_angle can influence this values (PI == ServerOptions::visible_angle * PI/90.0);
  */
  static Angle view_angle_width_rad();

  /** calculates my distance to another object. */
  static Value my_distance_to(Vector target);
  /** calculates my distance to the ball. */
  static Value my_distance_to_ball();
  /** calculates my distance to one of my teammates. */
  static Value my_distance_to_teammate(int number);
  /** calculates my distance to one of my opponents. */
  static Value my_distance_to_opponent(int number);
  /** calculates my angle to another object. */
  static Angle my_angle_to(Vector target);
  /** calculates my neck angle to another object. */
  static Angle my_neck_angle_to(Vector target);
  /** checks whether the ball can be kicked.*/
  static Bool is_ball_kickable();
  /** checks whether the ball can be sensed.*/
  static Bool is_ball_infeelrange();
  /** checks whether the ball can be kicked (same as is_ball_kickable()).*/
  static Bool is_ball_kickable_exact();
  /** checks whether the goalie can catch the ball
      note the ban_cycle is not considered, so please do it in your policy 
  */
  static Bool is_ball_catchable();
  inline static Bool is_ball_kickable_for_me() { return is_ball_kickable(); }
  inline static Vector opponent_goalpos(){return Vector(ServerOptions::pitch_length/2.,0);};


  /** Returns if or if not ball is kickable in the next simulator step, when no kick command effects it. */
  static Bool is_ball_kickable_next_time();
  static Bool was_ball_kickable_last_time();
  static Bool is_ball_infeelrange_next_time();
  static Bool is_ball_outfeelrange_next_time();


  // 2. Functions for the ball:
  /** checks whether the ball position has been accurately determined. */
  static Bool is_ball_pos_valid();
  /** checks whether the ball is inside of the pitch.*/
  static Bool is_ball_in_pitch();
  /** gets the absolute ball position. */
  static Vector ball_pos_abs();
  /** Returns the position of the ball in the next simulator step, if no kick command affects it. */
  static Vector ball_next_time_pos_abs();
  /** gets the ball position relatively to me. */
  static Vector ball_pos_rel();
  /** gets the absolute ball velocity. */
  static Vector ball_vel_abs();
  /** gets the ball velocity relatively to me. */
  static Vector ball_vel_rel();
  /** checks whether the ball velocity has been accurately determinde. */
  static Bool is_ball_vel_valid();
  /** calculates the distance of the ball to another object. */
  static Value ball_distance_to(Vector target);

  // 3. Functions for the player's teammates:
  /** gets the index in the array at MDPstate for the teammate with uniform number 'number'.*/
  static int teammate_array_index(int number); 
  /** get the teammate that currently controls the ball */
  static int get_teammate_controlling_ball(Value kickrange_factor = 2.0); 
  /** checks whether the position of a teammate has been accurately determined. */
  static Bool is_teammate_pos_valid(int number);
  /** gets the absolute position of a teammate. */
  static Vector teammate_pos_abs(int number);
  /** checks whether the velocity of a teammate has been accurately determined. */
  static Bool is_teammate_vel_valid(int number);
  /** gets the absolute velocity of a teammate. */
  static Vector teammate_vel_abs(int number);
  /** checks whether a teammante is in an offside position. */
  static Bool is_teammate_offside(int number);
  /** gets the distance of a teammate to another object. */
  static Value teammate_distance_to(int number, Vector taret);
  /** gets the distance of a teammate to the ball .*/
  static Value teammate_distance_to_ball(int number);
  /** gets the distance of the teammate to me. */
  static Value teammate_distance_to_me(int number);
  /** gets the closest teammate to an object. */
  static int teammate_closest_to(Vector target);
  /** gets the closest teammate to the ball. */
  static int teammate_closest_to_ball();
  /** gets the closest teammate to me */
  static int teammate_closest_to_me();
  static int teammate_closest_to_ball_wme();
  /** gets the closest teammate player to the ball (without the goalie) */
  static int teammate_closest_to_ball_wme_wogoalie();
  /** checks whether the ball is kickable for a teammate. */
  static Bool is_ball_kickable_for_teammate(int number);
  //here a closest function with myself
  static int teammate_closest_to_wme(Vector target);
  //x-pos of last field player
  static Value last_player_line();


  // 4. Functions for the player's opponents:
  /** gets the index in the array at MDPstate for the opponent with uniform number 'number'.*/
  static int opponent_array_index(int number); 
  /** checks whether the position of an opponent has been accurately determined. */
  static Bool is_opponent_pos_valid(int number);
  /** gets the absolute position of an opponent. */
  static Vector opponent_pos_abs(int number);
  /** gets the absolute velocity of a teammate. */
  static Vector opponent_vel_abs(int number);
  /** checks whether a teammante is in an offside position. */
  static Bool is_opponent_offside(int number);
  /** checks whether an opponent is in my kickrange. */
  static Bool is_opponent_in_kickrange();
  /** gets the distance of an opponent to another object. */
  static Value opponent_distance_to(int number, Vector taret);
  /** gets the distance of an opponent to the ball .*/
  static Value opponent_distance_to_ball(int number);
  /** gets the distance of the opponent to me. */
  static Value opponent_distance_to_me(int number);
  /** gets the number of the closest opponent to an object DOES NOT CONSIDER FELT PLAYERS!*/
  static int opponent_closest_to(Vector target);
  /** gets the closest opponent to an arbitrary target, returns op. dist and dir from target */
  static int opponent_closest_to(Vector target,Value &dist,Value &dir);
  /** gets the closest opponent to an arbitrary target, returns op. dist and dir from target */
  static int opponent_closest_to(Vector target,Value &dist,Value &dir,Vector &pos);
  /** gets the closest opponent to the ball. */
  static int opponent_closest_to_ball(Value &dist, Value &dir, Vector &pos);
  /** gets the closest opponent to the goal - probably the goalie */
  static int opponent_closest_to_opponent_goal();
  /** gets the goalie age - if goalie not known -1 is returned */
  static int opponent_goalie_age();
  /** gets the goalie number - if known - ok. if not, take closest op. to gaol */
  static int opponent_goalie_number();
  /** gets the distance to the opponent goalie */
  static Value opponent_goalie_dist();
  /** gets the distance to the opponent goalie */
  static Value opponent_goalie_dist(Value &dir);
  /** gets the closest opponent to me */
  /* should not be used any more, because number might be invalid!!!
  static int opponent_closest_to_me();
  */
  /** gets the closest opponent to me, also returns position */
  static int opponent_closest_to_me(Vector &pos);
  /** gets the closest opponent to me, returns distance and direction */
  static int opponent_closest_to_me(Value &dist, Value &dir);
  /** gets the closest opponent to me, returns distance and direction and absolute position*/
  static int opponent_closest_to_me(Value &dist, Value &dir, Vector &pos);
  /** checks whether the ball is kickable for an opponent. */
  static Bool is_ball_kickable_for_opponent(int number);
  /** gets the number of opponents within radius of position, 
      it saves up to max_size postions of opponents in this range */
  static int opponents_within_range(Vector position, Value radius,int max_size= 0, Vector * opponents_pos=0);
  /** special: gets the number of opponents within my 'feel-range'.
      If mdp.number_of_sensed_opponents is set, then this value is returned,
      otherwise the procedure opponents_within_range() is called. */
  static int opponents_within_feelrange();
  /** gets the number of opponents within area p1 (left lower), p2 (right upper) */
  static int opponents_within_area(Vector p1, Vector p2);
  /** gets the number of opponents within area p1 (left lower), p2 (right upper) */
  static int opponents_behind_line(Value x);

  // 5a. functions to get knowledge of the goalie numbers:
  /** return our goalies player number */
  static int our_goalie_number() {return 1;}

  /** return opponent goalies player number */
  static int their_goalie_number(){return opponent_goalie_number();}

  // 5b. Common functions for all players:

  /** computes the distance of point from line between start and end 
   returns -1 if distance between line and line is not vertical */
  static float get_distance_from_line(Vector start, Vector end, Vector point);

  static float get_op_dist2line(Vector a, Vector b);

  /** gets the absolute position of an object .*/
  static Vector get_pos_abs(FPlayer player);

  /** gets the age of a player position .*/
  static int teammate_age(int number);
  /** gets the age of player position .*/
  static int age_playerpos(FPlayer *player);
  /** gets the age of player velocity .*/
  static int age_playervel(FPlayer *player);
  /** gets the age of ball position .*/
  static int age_ballpos();
  /** gets the age of ball velocity .*/
  static int age_ballvel();
  /** gets a player object inside of an area. 
   (Please add further doc here, Ralf...) */
  static const FPlayer *get_a_player_in_area(Value dist ,Angle angle,Value of_dist,Angle of_angle );
  /** gets the number of players in an area. */
  static int number_of_players_in_area(Value dist ,Angle angle,Value of_dist = 0,Angle of_angle = 0);

  // 6. Functions to change representations (e.g. relative to absolute)
  /** convert a position from polar to vector representation. */
  static Vector polar2global_pos(Value distance, Angle angle);
  /** checks whether a position is located inside of the pitch. */
  static Bool is_position_in_pitch(Vector position, float safety_margin = 0.0);

  // 7. Fuctions which treat properties of whole teams
  /** ??? ask Ralf. */
  static Value my_team_pos_of_offside_line();
  /** ??? ask Ralf. */
  static Value his_team_pos_of_offside_line();
  /** checks whether my team is in an offside position .*/
  static Bool is_my_team_offside();
  /** checks whether the opponent team is in an offside position. */
  static Bool is_his_team_offside();
  /**determines which team got the player who is fastest to ball (ART's method!) */
  static int fastest_team_to_ball();
  /**determines which team is fastest to ball with momentum */
  static int fastest_team_to_ball_momentum();
  /**determines whether my team is attacking */
  static bool is_my_team_attacking();
  /**determine who was last at ball of our team*/
  static int teammate_last_at_ball();

  // 8. Functions who check positions of objects.
  /** check whether an object is in our penalty area.*/
  static bool is_object_in_my_penalty_area(Vector obj);
  /** check whether an object is in their penalty area.*/
  static bool is_object_in_their_penalty_area(Vector obj);

  //9. mdp related get functions.
  /** gets our teamname.*/
  static char* team_name();
  /** gets our side.
      @return SIDE_LEFT (1) if we are playing from the left
      | SIDE_RIGHT (-1) if we are playing from the right
  */
  //static int my_side();

  /** get self.
      @return pointer to the player himself
  */
 static FPlayer* me();
  /** get current time. 
      @return int current timestep
  */
 static int time_current();
  /** get time of last update.
      @return int time of last update
  */
 static int time_of_last_update();
  /** computes the number of steps that are possible if player dashes with power 
      avoiding that stamina goes below recover_dec_thr * stamina_max
      if no restrictions apply then mdpInfo::stamina4infinite_steps is returned
      @return int number of steps that are possible
  */
 static int stamina4steps_if_dashing(Value power);
 /** computes the number of meters that are possible if player dashes with power 
      uses stamina4steps_if_dashing()
 */
 static Value stamina4meters_if_dashing(Value power);
  /** computes the available Power regarding recoverz threshold
  */
 static Value stamina_left();
  /** get the team identificator of who was last at the ball.
      @return int UNKNOWN_TEAM=0 | MY_TEAM=1 | HIS_TEAM=2
  */
 static int team_last_at_ball(); //Werte aus dem Bereich MDPstate::MY_TEAM,HIS_TEAM
  /** get the current playmode */
 static int play_mode();
  /** get my team score */
 static int my_team_score();
  /** get his team score */
 static int his_team_score();
   // ... more information values
  /** get ball information.
   @return a pointer to the ball
  */
 static FBall* ball();
  /** get a player from our team.
     @param player_nbr (range = 0-10)
     @return a pointer to a player
  */
 static FPlayer* my_team(int player_nbr);
  /** get a player from their team.
     @param player_nbr (range = 0-10)
     @return a pointer to a player
  */
 static FPlayer* his_team(int player_nbr);


 /* stuff concerning the computation of the current game state */
 /* return the game_state from mdp state */
 /* Game States*/
#define gsUNDEFINED -1
#define gsDEFENSE 0
#define gsBUILDUP 1
#define gsATTACK 2

 static int get_game_state();

 /* update game_state in mdp state (return true if game_state changes)*/
 static bool update_game_state(int myteam_steps, int opponent_steps);
  
 /* condition for transition from DEFENSE to BUILDUP*/
 static bool transistion_DB(Value ball_line, int myteam_steps, int opponent_steps); 
 /* condition for transition from BUILDUP to DEFENSE*/
 static bool transistion_BD(Value ball_line, int myteam_steps, int opponent_steps); 
 /* condition for transition from BUILDUP to ATTACK*/
 static bool transistion_BA(Value ball_line, int myteam_steps, int opponent_steps); 
 /* condition for transition from ATTACK to BUILDUP*/
 static bool transistion_AB(Value ball_line, int myteam_steps, int opponent_steps); 

/*  private: */
/*   mdpInfo(){} */
/*   ~mdpInfo(){} */
};

#endif
