/** \file base_moves.h
    The base move interfaces are defined here
    What is a move ?
    A move is an abstraction of a command. It makes it possible to execute
    several commands in a row and thus define complex commands like neuro_kick, 
    intercept_ball etc...
    Please name all files in which moves are defined move_xxx.[hc].
    Please put each move in a separate file, except if they are very small and 
    have something in common.
*/

#ifndef _BASE_MOVES_H_
#define _BASE_MOVES_H_

#include <iostream.h>
#include "cmd.h"
#include "move_defs.h"


class Abstract_Move{
friend class Move_Factory;
 protected:
  bool in_use;
  /** max_steps is a counter which is initialized to a positive value and 
      which is decremented after each step until it reaches 0 (end of move)
  */
  int max_steps;
  int steps_done;

  /** move statistics data structure
  */
  struct {
    int called;
    int succeeded;
    int failed;
    float avg_costs;
  } statistics;

 
  /** execute the move.
      Overload it to create your own moves/skills.
  */
  virtual void execute(Cmd &)=0;
  /** The final state of a move is to be defined here.
      Overload this method to implement your own move termination conditions.
      Simply return false if you want the move to be aborted after max_steps steps.
   */
  virtual bool final_state()=0;
 public:
  /** get the unique id of the move.
      \return move id.
  */ 
  inline virtual int id() {return ABSTRACT_MOVE;}
  /** get a short description of the move */
  virtual const char* id_str() { return "??? no name ???"; }
  /** check if the move is finished.
      A move is finished if it has reached its final state or max_steps == 0
      The move is also given free for the fabric if it is finished.
  */
  virtual bool finished();
  /** execute the move and decrease max_steps by one. */
  virtual void step(Cmd &);
  /** reset the move. This forces a new decision */
  inline virtual void reset(int new_steps=0) {max_steps = new_steps;}

  /** terminate the move. after calling terinate(), for the move itself and for all
      submoves of the move, the following conditions have to be fullfilled:
      1. max_steps = 0
      2. in_use = false
      3. steps_done = 0
      Attention: if your move uses submoves you have to overload terimate() to terminate them! 	    
  */
  inline virtual void terminate(){max_steps=0;in_use=false;}

  /** resets the move statistics data structure 
   */
  inline void reset_statistics(){statistics.called=statistics.succeeded=statistics.failed=0;statistics.avg_costs=0;}
  
  /** prints the move statistics data structure 
   */
  void print_statistics();
};


/** the base class for main moves (returning dash, kick, turn, catch) */
class Main_Move : public Abstract_Move{
friend class Move_Factory;
protected:
  /**
   */
  Main_Move() { in_use= false; max_steps= 0; steps_done = 0; reset_statistics();}

  /** execute the move.
      Overload it to create your own moves/skills.
  */
  virtual void execute(Cmd &) {}
  /** The final state of a move is to be defined here.
      Overload this method to implement your own move termination conditions.
      Simply return false if you want the move to be aborted after max_steps steps.
   */
  virtual bool final_state() { return true; }
public:
  /** get the unique id of the move.
      \return move id.
  */ 
  inline virtual int id() {return MAIN_MOVE;}
};

/** the base class for neck moves (returning turn_neck) */ 
class Neck_Move : public Abstract_Move{
friend class Move_Factory;
protected:
  /**
   */
  Neck_Move() { in_use= false; max_steps= 0; }

  /** execute the move.
      Overload it to create your own moves/skills.
  */
  virtual void execute(Cmd &) {}
  /** The final state of a move is to be defined here.
      Overload this method to implement your own move termination conditions.
      Simply return false if you want the move to be aborted after max_steps steps.
   */
  virtual bool final_state() { return true; }
public:
  /** get the unique id of the move.
      \return move id.
  */ 
  inline virtual int id() {return NECK_MOVE;}
};

/** the base class for say moves (returning say) */ 
class Say_Move : public Abstract_Move{
friend class Move_Factory;
protected:
  /**
   */
  Say_Move() { in_use= false; max_steps= 0; }

  /** execute the move.
      Overload it to create your own moves/skills.
  */
  virtual void execute(Cmd &) {}
  /** The final state of a move is to be defined here.
      Overload this method to implement your own move termination conditions.
      Simply return false if you want the move to be aborted after max_steps steps.
   */
  virtual bool final_state() { return true; }
public:
  /** get the unique id of the move.
      \return move id.
  */ 
  inline virtual int id() {return SAY_MOVE;}
};

/** the base class for view moves (returning change_view) */ 
class View_Move : public Abstract_Move{
friend class Move_Factory;
protected:
  /**
   */
  View_Move() { in_use= false; max_steps= 0; }

  /** execute the move.
      Overload it to create your own moves/skills.
  */
  virtual void execute(Cmd &) {}
  /** The final state of a move is to be defined here.
      Overload this method to implement your own move termination conditions.
      Simply return false if you want the move to be aborted after max_steps steps.
   */
  virtual bool final_state() { return true; }
public:
  /** get the unique id of the move.
      \return move id.
  */ 
  inline virtual int id() {return SAY_MOVE;}
};


#endif // _BASE_MOVES_H_

