/** \file move_basics.h
    All moves that simply represent an Action are defined here.
*/

#ifndef _MOVES_BASICS_H_
#define _MOVES_BASICS_H_

#include <iostream.h>
#include <stdio.h>

#include "move_defs.h"
#include "base_moves.h"
#include "cmd.h"
#include "types.h"

//#include "globals.h"

class Move_Dash : public Main_Move {
friend class Move_Factory;
protected:
  double power;
  /** returns a dash action */
  inline void execute(Cmd & cmd){cmd.cmd_main.set_dash(power);}
  /** The final state of a move is to be defined here.
      Overload this method to implement your own move termination conditions.
      Simply return false if you want the move to be aborted after max_steps steps.
   */
  inline virtual bool final_state(){return false;}

public:
  /** constructor */
  Move_Dash(double p_power=30);
  /** get the unique id of the move.
      \return move id.
  */ 
  inline int id() {return MOVE_DASH;}
  /** get a short description of the move */
  const char* id_str() { return "Move_Dash"; }
};

class Move_Kick : public Main_Move{
friend class Move_Factory;
protected:
  double power;
  double direction;
  /** The final state of a move is to be defined here.
      Overload this method to implement your own move termination conditions.
      Simply return false if you want the move to be aborted after max_steps steps.
   */
  inline virtual bool final_state(){return false;}
  /** returns a kick action */
  inline void execute(Cmd & cmd){cmd.cmd_main.set_kick(power, direction);}
public:
  /** constructor.
      \param p_power power (range 0-100)
      \param P_direction direction (range -180 ~ 180 degrees)
  */
  Move_Kick(double p_power=30, double p_direction=0);
  /** get an id for the action */
  inline int id() {return MOVE_KICK;}
  /** get a short description of the move */
  const char* id_str() { return "Move_Dash"; }
};

/** Catch Move */
class Move_Catch : public Main_Move{
friend class Move_Factory;
protected:
  double direction;
  /** The final state of a move is to be defined here.
      Overload this method to implement your own move termination conditions.
      Simply return false if you want the move to be aborted after max_steps steps.
   */
  inline virtual bool final_state(){return false;}
  /** returns a catch action */
  inline void execute(Cmd & cmd){cmd.cmd_main.set_catch(direction);}
public:
  /** constructor.
      \param p_direction direction (range -180 ~ 180 degrees)
  */
  Move_Catch(double p_direction=0);
  /** get an id for the action */
  inline int id() {return MOVE_CATCH;}
  /** get a short description of the move */
  const char* id_str() { return "Move_Catch"; }
};

/** Turn Move */
class Move_Turn : public Main_Move{
friend class Move_Factory;
protected:
  double moment;
  /** The final state of a move is to be defined here.
      Overload this method to implement your own move termination conditions.
      Simply return false if you want the move to be aborted after max_steps steps.
   */
  inline virtual bool final_state(){return false;}
  /** returns a turn action */
  inline void execute(Cmd & cmd){cmd.cmd_main.set_turn(moment);}
public:
  /** constructor.
      \param p_moment moment (range -180 ~ 180 degrees)
  */
  Move_Turn(double p_moment=60);
  /** get an id for the action */
  inline int id() {return MOVE_TURN;}
  /** get a short description of the move */
  const char* id_str() { return "Move_Turn"; }
};

/** Move Move */
class Move_Move : public Main_Move{
friend class Move_Factory;
protected:
  double x;
  double y;
  /** The final state of a move is to be defined here.
      Overload this method to implement your own move termination conditions.
      Simply return false if you want the move to be aborted after max_steps steps.
   */
  inline virtual bool final_state(){return false;}
  /** returns a move action */
  inline void execute(Cmd & cmd){cmd.cmd_main.set_moveto(x, y);}
public:
  /** constructor.
      \param p_x x coordinate (range -54 ~ 54)
      \param p_y y coordinate (range -32 ~ 32)
      
  */
  Move_Move(double p_x=-23, double p_y=23);
  /** get an id for the action */
  inline int id() {return MOVE_MOVE;}
  /** get a short description of the move */
  const char* id_str() { return "Move_Move"; }
};

/** Turn Neck Move */
class Move_Turn_Neck : public Neck_Move{
friend class Move_Factory;
protected:
  double moment;
  /** The final state of a move is to be defined here.
      Overload this method to implement your own move termination conditions.
      Simply return false if you want the move to be aborted after max_steps steps.
   */
  inline virtual bool final_state(){return false;}
  /** returns a turn_neck action */
  inline void execute(Cmd & cmd){if(!cmd.cmd_neck.is_lock_set()) 
    cmd.cmd_neck.set_turn(moment);}
public:
  /** constructor.
      \param p_moment moment (range -180 ~ 180 degrees)
  */
  Move_Turn_Neck(double p_moment=0);
  /** get an id for the action */
  inline int id(){return MOVE_TURN_NECK;}
  /** get a short description of the move */
  const char* id_str() { return "Move_Turn_Neck"; }
};

/** Say Move */
class Move_Say : public Say_Move{
friend class Move_Factory;
protected:
  char *message;
  int priority;
  /** The final state of a move is to be defined here.
      Overload this method to implement your own move termination conditions.
      Simply return false if you want the move to be aborted after max_steps steps.
   */
  inline virtual bool final_state(){return false;}
  /** returns a say action */
  inline void execute(Cmd & cmd){cmd.cmd_say.set_message(message,priority);}
public:
  /** constructor.
      \param p_mesage message
  */
  Move_Say(char *p_message="");
  /** get an id for the action */
  inline int id(){return MOVE_SAY;}
  /** get a short description of the move */
  const char* id_str() { return "Move_Say"; }
};

/** Change the View Move. */
class Move_Change_View: public View_Move{
friend class Move_Factory;
protected:
  int view_width;
  int view_quality;
  /** The final state of a move is to be defined here.
      Overload this method to implement your own move termination conditions.
      Simply return false if you want the move to be aborted after max_steps steps.
   */
  inline virtual bool final_state(){return false;}
  /** returns a change_view action */
  inline void execute(Cmd & cmd){cmd.cmd_view.set_angle_and_quality(view_width, view_quality);}
public:
  /** constructor.
      \param p_width view_width ( NARROW | NORMAL | WIDE )
      \param p_quality view_quality ( HIGH | LOW )
  */
  Move_Change_View(int p_width=1, int p_quality=1);
  /** get an id for the action */
  inline int id(){return MOVE_CHANGE_VIEW;}
  /** get a short description of the move */
  const char* id_str() { return "Move_Change_View"; }
};

/**
   this move just triggers a single turn command but adjusts the turn moment to the current
   velocity (according to the interia physics defined in the soccer server, see server manual 
   chapter 3 for more details). this means, that move_turn_inert will result in a turn with 
   the specified moment independent of the players velocity

   in newBS99 this functionality was delivered by the manipulate_turn pocedure, which was called
   inside the behave function and manipulated  every turn command returned by a move automatically. 
   if you want your moves to use inertia respected turns in bs2k just use the Move_Turn_Inertia instead
   of setting turn commands in the command form.
*/

class Move_Turn_Inertia : public Main_Move {
  friend class Move_Factory; 
 protected:
  double moment;

  /** dont care about final states in single command moves */
  inline virtual bool final_state(){return false;}

  /** sets a velocity adjusted turn command */
  void execute(Cmd & cmd);

public:
  /** constructor.
      \param p_moment moment (range -180 ~ 180 degrees)
  */
  Move_Turn_Inertia(double p_moment=60);
  /** get an id for the action */
  inline int id() {return MOVE_TURN_INERTIA;}
  /** get a short description of the move */
  const char* id_str() { return "Move_Turn_Inertia"; }
}; 

/**
   this move just triggers a single dash command but adjusts the requested dash power to save stamina.
   there are different scenarios where one can save stamina:

   (1) if the enemy has a free_kick, kick_in or something like that its not allowed to our players
       to get closer to the ball than SP_FREE_KICK_DISTANCE. So we should prevent our player from 
       dashing insdide the this free kick area and being reseted to the border of this area by the 
       referee immediately because this is a great waste of stamina
   (2) if a dash with the requested power would lead to a velocity greater than Player_Max_Speed and
       therefore will be cut of a smaller power value is sufficient for the same effect but saves stamina
   (3) the expected stamina value os to small after a dash with the requestet power.

   in newBS99 this functionality was delivered by the manipulate_dash pocedure which was called
   inside the behave function and manipulated  every dash command returned by moves automatically. 
   if you want your moves to use stamina respected dashes in bs2k just use the Move_Dash_Stamina instead
   of Move_Dash or setting dash commands in the command form manually.
*/
class Move_Dash_Stamina : public 
Abstract_Move {
  friend class Move_Factory; 
 protected:
  double power;
 
  /** dont care about final states in single command moves */
  inline virtual bool final_state(){return false;}
 
  /** sets a stamina adjusted dash command */
  void execute(Cmd & cmd);

public:
  /** constructor.
      \param p_moment moment (range -180 ~ 180 degrees)
  */
  Move_Dash_Stamina(double p_power=60);

  /** get an id for the action */
  inline int id() {return MOVE_DASH_STAMINA;}
  /** get a short description of the move */
  const char* id_str() { return "Move_Dash_Stamina"; }

}; 







#endif // _MOVES_BASICS_H_
