/** \file move_factory.h
    This is the header file for an object management class. 
*/

#ifndef _MOVE_FACTORY_H_
#define _MOVE_FACTORY_H_

#include "move_basics.h"
#include "types.h"
#include <iostream.h>
#include "move_1step_kick.h"
#include "move_run_to.h"
#include "move_intercept_ball.h"
//#include "move_stop_ball.h"
#include "move_look_around.h"
#include "move_face_ball.h"

#include "neuro_kick.h"
#include "neuro_carry_ball.h"
#include "neuro_intercept_ball.h"
#include "neuro_kick2.h"
#include "neuro_stop_ball.h"
#include "neuro_attack.h"
#include "neuro_go2pos.h"
#include "neuro_hold_ball.h"
#include "neuro_dribble.h"
#include "neuro_dribble2.h"

#include "move_go2pos_backwards.h"

#include "move_goalie_ellipsoid.h"
#include "neuro_goalie_intercept.h"

/** Move Factory.
    A Factory is a class which manages the creation/resources of certain objects.
    In this case there is always one instance of each Move which is passed
    by reference to the caller.
*/
class Move_Factory{
private:
  /* Add a pointer to your new move here. Do not forget to initialize it in the constructor. */
  static Move_Kick *_kick;
  static Move_Turn *_turn;
  static Move_Dash *_dash;
  static Move_Catch *_catch;
  static Move_Move *_move;
  static Move_Say *_say;
  static Move_Turn_Neck *_turn_neck;
  static Move_Change_View *_change_view;

  static Main_Move *_main_move;
  static Neck_Move *_neck_move;
  static Say_Move  *_say_move;
  static View_Move *_view_move;

  static Move_Turn_Inertia *_turn_inertia;
  static Move_Dash_Stamina *_dash_stamina;

  static Move_1Step_Kick *_1step_kick;

  /** converted newBS99 moves/skills */
  static Move_Run_To *_run_to;
  static Move_Intercept_Ball *_intercept_ball;  
  static Move_Look_Around *_look_around;
  static Move_Face_Ball *_face_ball;

  static Move_Go2Pos_Backwards *_go_backwards;

  /** neuro moves */
  static Neuro_Kick *_neuro_kick;
  static Neuro_Carry_Ball *_neuro_carry_ball;
  static Neuro_Intercept_Ball *_neuro_intercept_ball;
  static Neuro_Kick2 *_neuro_kick2;
  static Neuro_Stop_Ball *_neuro_stop_ball;
  static Neuro_Attack *_neuro_attack;
  static Neuro_Go2Pos *_neuro_go2pos;
  static Neuro_Hold_Ball *_neuro_hold_ball;
  static Neuro_Dribble *_neuro_dribble;
  static Neuro_Dribble2 *_neuro_dribble2;

  static float shoot2goal_vel; // default velocity for goalshots
  /* goalie moves */
  static Move_Ellipsoid *_move_ellipsoid;
  static Neuro_Goalie_Intercept *_neuro_goalie_intercept;

public:
  static void init();
  static void print_statistics();
  /** pseudo-constructor.
      _main_move is a dummy main move, 
      call this method if you wish a move with no functionaliy and
      wich is "finished()" immediatly.
   */
  static Main_Move * get_Main_Move(int p_max_steps=1) {  return _main_move; }
  /** pseudo-constructor.
      _neck_move is a dummy neck move, 
      call this method if you wish a move with no functionaliy and
      wich is "finished()" immediatly.
   */
  static Neck_Move * get_Neck_Move(int p_max_steps=1) {  return _neck_move; }
  /** pseudo-constructor.
      _say_move is a dummy say move, 
      call this method if you wish a move with no functionaliy and
      wich is "finished()" immediatly.
   */
  static Say_Move * get_Say_Move(int p_max_steps=1) {  return _say_move; }
  /** pseudo-constructor.
      _view_move is a dummy view move, 
      call this method if you wish a move with no functionaliy and
      wich is "finished()" immediatly.
   */
  static View_Move * get_View_Move(int p_max_steps=1) {  return _view_move; }

  /** pseudo-constructor.
      \param p_power power (range 0-100)
      \param P_direction direction (range -180 ~ 180 degrees)
  */
  static Main_Move * get_Kick(double p_power=30, double p_direction=0, int p_max_steps=1);

  /** pseudo-constructor.
      \param p_direction direction (range -180 ~ 180 degrees)
  */
  static Main_Move * get_Catch(double p_direction=0, int p_max_steps=1);
  /** pseudo-constructor.
      \param p_moment moment (range -180 ~ 180 degrees)
  */
  static Main_Move * get_Turn(double p_moment=60, int p_max_steps=1);

  /** pseudo-constructor.
      \param p_x x coordinate (range -54 ~ 54)
      \param p_y y coordinate (range -32 ~ 32)
      
  */
  static Main_Move * get_Move(double p_x=-23, double p_y=23, int p_max_steps=1);
  /** pseudo-constructor.
      Please note that negative power uses \b double stamina.
      \param p_power power (range -100 ~ 100)
  */
  static Main_Move * get_Dash(double p_power=30, int p_max_steps=1);

  /** pseudo-constructor.
      \param p_moment moment (range -180 ~ 180 degrees)
  */
  static Neck_Move * get_Turn_Neck(double p_moment=0, int p_max_steps=1);
  /** pseudo-constructor.
      \param p_mesage message
      \param priority sending priority 0 = highest
  */
  static Say_Move * get_Say(char *p_message="", int p_priority = 0, int p_max_steps=1);
  /** pseudo-constructor.
      \param p_width view_width ( NARROW | NORMAL | WIDE )
      \param p_quality view_quality ( HIGH | LOW )
  */
  static View_Move * get_Change_View(int p_width=Cmd_View::VIEW_ANGLE_NORMAL, int p_quality=Cmd_View::VIEW_QUALITY_HIGH, int p_max_steps=1);
 
  /** pseudo-constructor.
      \param p_moment moment (range -180 ~ 180 degrees)
  */
  static Main_Move * get_Turn_Inertia(double p_moment=60, int p_max_steps=1);

  /** pseudo-constructor.
      Please note that negative power uses \b double stamina.
      \param p_power power (range -100 ~ 100)
  */
  static Main_Move * get_Dash_Stamina(double p_power=30, int p_max_steps=1);


  /** pseudo-constructor.
   */
  static Main_Move * get_1Step_Kick(Value vel,Angle dir);
  /** pseudo-constructor.
      \param p_x x coordinate (range -54 ~ 54)
      \param p_y y coordinate (range -32 ~ 32)
   */
  static Main_Move * get_Run_To(double p_x=-30, double p_y=0, double p_accuracy=2.0, int p_max_steps=1);

  /** pseudo-constructor.
    \param p_safety_steps number of bonus steps to meet the rendevouz point before the ball     
  */
  static Main_Move * get_Intercept_Ball(int p_safety_steps=1, int p_max_steps=1);
 
  /** pseudo-constructor.
      \param p_velocity_threshold    
  */
  //static Main_Move * get_Stop_Ball(double p_velocity_threshold=1.0, int p_max_steps=1);
 
  /** pseudo-constructor.
   */
  static Main_Move * get_Look_Around(int p_max_steps=5);

  /** pseudo-constructor.
      \param p_max_steps max steps
  */
  static Main_Move * get_Face_Ball(int p_max_steps=40);

  /** pseudo-constructor.
      @alex What does this mean ???
      \param p_target_velocity wanted ball velocity (range 0 ~ 2.7)
      \param p_target_direction wanted ball direction (range 0 ~ 2PI)
  */
  static Main_Move * get_Neuro_Kick(int mode = 0, double p1 =2.0, double p2 = 0.0, double p3 = 0.0, int p_max_steps=10);
  
  /** pseudo-constructor.
      @alex What does this mean ???
      \param p_target_velocity wanted ball velocity (range 0 ~ 2.7)
      \param p_target_direction wanted ball direction (range 0 ~ 2PI)
  */
  static Main_Move * get_Neuro_Kick2(int mode = 0, double p1 =2.0, double p2 = 0.0, double p3 = 0.0, int p_max_steps=10);
  
  /** pseudo-constructor.
      \param p_target_x 
      \param p_target_y 
  */
  static Main_Move * get_Neuro_Carry_Ball(double p_target_x=0.0, double p_target_y=0.0, int p_max_steps=25);


  /** pseudo-constructor.
  */
  static Main_Move * get_Neuro_Intercept_Ball(int p_max_steps=5);
 
  /** pseudo-constructor.
  */
  static Main_Move * get_Neuro_Stop_Ball(float p_max_velocity = 0.7, int p_max_steps=5);
  
  /** pseudo-constructor.
  */
  static Main_Move * get_Neuro_Hold_Ball(int p_max_steps=5);
  
  /** pseudo-constructor.
  */
  static Main_Move * get_Neuro_Hold_Ball_Turn(int p_max_steps=5);
 
  /** pseudo-constructor.
  */
  static Main_Move * get_Neuro_Hold_Ball_Look(float ang, int p_max_steps=5);
  
  /** pseudo-constructor.
   */
  static Main_Move * get_Neuro_Attack(int p_max_steps=10);
 
  /** pseudo-constructor.
   */
  static Main_Move * get_Neuro_Dribble(double p_target_x, double p_target_y, int p_max_steps=5);
 
  /** pseudo-constructor.
   */
  static Main_Move * get_Neuro_Dribble2(double p_target_x, double p_target_y, int p_max_steps=5);
   
  /** pseudo-constructor.
   */
  static Main_Move * get_Neuro_Go2Pos(float p_target_x=0, float p_target_y=0, int p_max_steps=5);
 
  /** Move neuro go 2 position with consideration of stamina
   */
  static Main_Move * get_Neuro_Go2Pos_stamina(float p_target_x=0, float p_target_y=0, float p_dash_power = 100, float p_min_dash_power = 40, float p_inc_dash_power = 20,int p_max_steps=5);
 
  /** Move neuro go 2 position with consideration of stamina and tolerance
   */
  static Main_Move * get_Neuro_Go2Pos_tolerance(float p_target_x=0, float p_target_y=0, float p_dash_power = 100, float p_min_dash_power = 40, float p_inc_dash_power = 20,float p_target_tolerance=1.0, int p_max_steps=5);
 

  /** Intercept_Ball value function evaluation intercface.
   */ 
  static Neuro_Intercept_Ball * get_Eval_Neuro_Intercept() {return _neuro_intercept_ball;}
  /** pseudo-constructor for Move_Ellipsoid.
      This move mokes the goalie position himself on an ellipsoid around the goal so that his angle to the ball is best.
     \param p_x_radius the x radius of the ellipsoid
     \param p_y_radius the y radius of the ellipsoid
  */
  static Main_Move *get_Move_Ellipsoid(double p_x_radius=4 , double p_y_radius=7, int p_max_steps=1);
 /** pseudo-constructor.
  */
  
  static Main_Move * get_Neuro_Goalie_Intercept(int p_max_steps=5);
  
  static Main_Move * get_Move_Go2Pos_Backwards(double p_x=-52, double p_y=0, double p_accuracy=1.0, double power=-50, int p_max_steps=3);

};

#endif // _MOVE_FACTORY_H_
