#include "neuro_attack.h"
#include "mdp_info.h"
#include "options.h"
#include "tools.h"
#include "move_factory.h"
#include "logger.h"

Neuro_Attack::Neuro_Attack(){
  
  char *netname = "./data/nets_neuro_attack/attack.net";
  
  // important: do init first, since net loading might be aborted
  nets[0] = new Net();
  if(nets[0]->load_net(netname) == FILE_ERROR){
    cout<<"Neuro_Attack: No net-file found in directory ./nets - stop loading\n";
    return;
  }
  net_counter = 1;

  /* set params */
  params.doku_level = 0;  
  params.costs_failure = 1.0;
  params.costs_success = 0.0;
   
  /* neuro intercept uses dashs */
  params.dash_power_min = 100;
  params.dash_power_inc = 10;
  params.dash_power_max = 100;
  
  /* neuro intercept uses turns */
  params.turn_angle_min = 2*PI/16;
  params.turn_angle_inc = 2*PI/16;
  params.turn_angle_max = 2*PI;
  
  /* neuro intercept uses no kicks */
  params.kick_power_min = 0;
  params.kick_power_inc = 0;
  params.kick_power_max = 0;

  params.kick_angle_min = 0;
  params.kick_angle_inc = 0;
  params.kick_angle_max = 0;  
 
 params.kick_finetune = 0;
  params.kick_finetune_power = 0;
  params.turn_finetune = 0;

  kickable_tolerance = 0.1;
}

/** destructor method */
Neuro_Attack::~Neuro_Attack(){
  delete nets[0];
}

bool Neuro_Attack:: decide_rule_based(Cmd & cmd){
  /* ridi: replaced
  int opponent = mdpInfo::opponent_closest_to_ball();
  current_state.opnt1_pos = mdpInfo::opponent_pos_abs(opponent);
  */
  Value dir,dist;
  int opponent=mdpInfo::opponent_closest_to_ball(dist,dir,current_state.opnt1_pos);
  
  LOG_MOVE << "Neuro_Attack: attacking opponent player " << opponent << " ." ;
  return true;
}

void Neuro_Attack::generate_actions(const Neuro_State& state){
  action_counter = 0;
  double power, angle;

  /* add turn commands to action set */
  Vector ball_dist = state.ball_pos - state.my_pos;
  float ball_view_angle = ball_dist.arg() - state.my_angle;
  for(angle=params.turn_angle_min; angle<params.turn_angle_max; angle+=params.turn_angle_inc){
    actions[action_counter].unset_lock();
    actions[action_counter].unset_cmd();
    
    float turn_angle = ball_view_angle+angle;
    turn_angle = Tools::get_angle_between_null_2PI(turn_angle);    
    actions[action_counter].set_turn(turn_angle);
    action_counter++;
  }

  /* add dash  commands to action set */
  for(power=params.dash_power_min; power<=params.dash_power_max; power+=params.dash_power_inc){    
    Neuro_Move::actions[action_counter].unset_lock();
    Neuro_Move::actions[action_counter].unset_cmd();
    
    Neuro_Move::actions[action_counter].set_dash(power);
    action_counter++;
  }
}

/** specifies the successful terminal states of the neuro intercept ball move

    intercept ball is finished successful if the ball is inside the 
    players kick range.
*/ 
bool Neuro_Attack::is_success(const Neuro_State& state){
  
  if(is_failure(state) == true){
    return false;
  }
  Vector ball_dist = state.my_pos - state.ball_pos;
  if(ball_dist.norm() <= ServerOptions::kickable_area-kickable_tolerance){
    //ball is in my kick range
    return true;
  }
  
  return false;
}

/** specifies the unsuccessful terminal states of the neuro intercept ball move

    there are no unsuccessful states for intercept ball
*/ 
bool Neuro_Attack::is_failure(const Neuro_State& state){
  return false;
}

/** features of neuro attack:

   
 */ 
void Neuro_Attack::get_features(const Neuro_State& state, float *net_in){
   
  Vector ball_dist = state.ball_pos - state.my_pos;
  Vector ball_vel = state.ball_vel;
  Vector my_vel = state.my_vel;

  Vector op_ball_dist = state.ball_pos - state.opnt1_pos;
  Vector op_my_dist = state.my_pos - state.opnt1_pos;
  float angle_diff = op_ball_dist.arg() - op_my_dist.arg();
  angle_diff = Tools::get_angle_between_mPI_pPI(angle_diff);


  float ball_angle = ball_dist.arg();
  float ball_view_angle = ball_dist.arg() - state.my_angle;  
  

  float my_op_angle = op_my_dist.arg() - state.my_angle;
  my_op_angle = Tools::get_angle_between_mPI_pPI(my_op_angle);


  ball_vel.rotate(2*PI-ball_angle);
  my_vel.rotate(2*PI-ball_angle);

  ball_view_angle = Tools::get_angle_between_mPI_pPI(ball_view_angle);
  
  net_in[0] = ball_dist.norm();
  net_in[1] = ball_view_angle;
  
  net_in[2] = my_vel.x;
  net_in[3] = my_vel.y;
  
  net_in[4] = ball_vel.x;
  net_in[5] = ball_vel.y;

  net_in[6] = angle_diff;
  net_in[7] = my_op_angle;
}


void Neuro_Attack::print_state(const Neuro_State& state){

  float features[6];
  get_features(state, features);
  
  printf("Attack step: %d:  ball_dist %.2f  ball_angle %.2f  ball_vel (%.2f %.2f) my_vel (%.2f %.2f) \n",
	 mdpInfo::mdp->time_current, features[0],  features[1],  features[4],  features[5], features[2],  features[3]);
}

bool Neuro_Attack::final_state(){
  int my_number = mdpInfo::mdp->me->number;
  
  /* terminate intercept if a teammate is able to kick the ball */
  int teammate =  mdpInfo::teammate_closest_to_ball();
  
  if(teammate != my_number){
    if(mdpInfo::is_ball_kickable_for_teammate(teammate)){
      //printf("player %d :kein intercept: player %d hat den ball...\n", my_number,  teammate);
      LOG_MOVE << "NeuroAttack: a teammate got the ball! Attack move terminated... ";
      return true;
    }
  }
  
  if(mdpInfo::my_distance_to_ball()<ServerOptions::kickable_area-kickable_tolerance){
    LOG_MOVE << "NeuroAttack: i got the ball! Attack move terminated... ";
    return true;
  }

  return false;           

}


