#include "neuro_carry_ball.h"
#include "mdp_info.h"
#include "options.h"
#include "tools.h"
#include "move_factory.h"
#include "logger.h"

#define MAX_TARGET_DIST 5.0

Neuro_Carry_Ball::Neuro_Carry_Ball(double p_target_x=0.0, double p_target_y = 0.0){
  target_x = p_target_x;
  target_y = p_target_y;
  
  last_time_executed = -1;
  
  char *netname = "./data/nets_neuro_carry_ball/carry_ball.net";
  
  // important: do init first, since net loading might be aborted
  nets[0] = new Net();
  if(nets[0]->load_net(netname) == FILE_ERROR){
    cout<<"NeuroCarryBall: No net-file found in directory ./nets - stop loading\n";
    return;
  }
  
  net_counter = 1;

  /* set documentation level */
  params.doku_level = 0;  
  params.costs_failure = 1.0;
  params.costs_success = 0.0;
  
  /* neuro carry ball uses dashs */
  params.dash_power_min = 25;
  params.dash_power_inc = 25;
  params.dash_power_max = 100;
  
  /* neuro carry ball uses turns */
  params.turn_angle_min = 2*PI/16;
  params.turn_angle_inc = 2*PI/16;
  params.turn_angle_max = 2*PI;

  /* neuro carry ball uses kicks */
  params.kick_power_min = 10;
  params.kick_power_inc = 10;
  params.kick_power_max = 50;

  params.kick_angle_min = 0;
  params.kick_angle_inc = 2*PI/16;
  params.kick_angle_max = 2*PI;
 
  params.kick_finetune = 0;
  params.kick_finetune_power = 0;
  params.turn_finetune = 0;

  /* set custom carry ball params */
  min_ball_dist = 0.1;
  target_tolerance = 2.0;
  max_target_velocity = 0.4;
   
  reset_statistics();
}

/** destructor method */
Neuro_Carry_Ball::~Neuro_Carry_Ball(){
  delete nets[0];
}

/** rule based pre-computing part of the neuro carry ball move  */ 
bool Neuro_Carry_Ball::decide_rule_based(Cmd & cmd){

  Vector target = Vector(target_x, target_y) -  mdpInfo::my_pos_abs();
  LOG_MOVE << "Neuro_Carry_Ball: Rel. target: (" << target.x << ", " << target.y << ")"; 
  int x1,x2,x3,x4;
  x1=(int)mdpInfo::my_pos_abs().x;
  x2=(int)mdpInfo::my_pos_abs().y;
  x3 = x1 + (int)target.x;
  x4 = x2 + (int)target.y;
  LOG_DRAWLINE<<x1<<" "<<x2<<" "<<x3<<" "<<x4<<" green";
    
  if(last_time_executed +1 != mdpInfo::mdp->time_current){
  }
  
  last_time_executed = mdpInfo::mdp->time_current;


  return true;
}

/** specifies the successful terminal states of the neuro carry ball move*/ 
bool Neuro_Carry_Ball::is_success(const Neuro_State& state){
  Vector target_area;
  target_area.x = target_x;
  target_area.y = target_y;

  Vector target_dist = state.my_pos-target_area;
  
  if(is_failure(state) == true){
    return false;
  }
  
  if((mdpInfo::is_ball_kickable() == true)
     &&(target_dist.norm() <= target_tolerance)
     &&(state.ball_vel.norm()<max_target_velocity)){
    //ball is in my Kickrange AND i am inside the target area AND ball vel is smaller than max.
    return true;
    
  }
  
  return false;
}

/** specifies the unsuccessful terminal states of the neuro carry ball move*/ 
bool Neuro_Carry_Ball::is_failure(const Neuro_State& state){

  Vector ball_dist = state.ball_pos - state.my_pos;
  if(ball_dist.norm() > 3.0){
    return true;
  }
  if(ball_dist.norm() < ServerOptions::player_size+min_ball_dist){
    return true;
  }  

  Vector target = Vector(target_x, target_y);
  float angle_diff = (target-state.my_pos).arg() - state.ball_vel.arg();
  angle_diff = Tools::get_angle_between_mPI_pPI(angle_diff);
  if(fabs(angle_diff) > 2*PI/8) {
    //LOG_MOVE_D << "Neuro_Carry_Ball: Be careful!";
    if(ball_dist.norm() > ServerOptions::kickable_area-0.2){
      return true;
    }
  }
  return false;
}

/** generate action set */
void Neuro_Carry_Ball::generate_actions(const Neuro_State& state){
  action_counter = 0;
  double power, angle;

  /* add turn commands to action set */
  for(angle=params.turn_angle_min; angle<params.turn_angle_max; angle+=params.turn_angle_inc){
    actions[action_counter].unset_lock();
    actions[action_counter].unset_cmd();
    
    actions[action_counter].set_turn(angle);
    action_counter++;
  }

  /* add dash  commands to action set */
  for(power=params.dash_power_min; power<=params.dash_power_max; power+=params.dash_power_inc){    
    actions[action_counter].unset_lock();
    actions[action_counter].unset_cmd();
    
    actions[action_counter].set_dash(power);
    action_counter++;
  }
 
  /* add kick commands to action set */
  if(mdpInfo::is_ball_kickable_exact() == true){

    for(power=params.kick_power_min; power<=params.kick_power_max; power+=params.kick_power_inc){
      for(angle=params.kick_angle_min; angle<params.kick_angle_max; angle+=params.kick_angle_inc){
	actions[action_counter].unset_lock();
	actions[action_counter].unset_cmd();
	
	actions[action_counter].set_kick(power, angle);
	action_counter++;
      }
    }
  }
  
}

bool Neuro_Carry_Ball::final_state(){
  Vector target_area;
  target_area.x = target_x;
  target_area.y = target_y;
  
  if(mdpInfo::my_distance_to_ball() > 3.) return true;

  Vector target_dist;
  target_dist.x = target_area.x - mdpInfo::mdp->me->pos_x.v;
  target_dist.y = target_area.y - mdpInfo::mdp->me->pos_y.v;
  
  if(target_dist.norm() < 2.0) {
    //printf("target reached!  %.2f %.2f\n", target_area.x, target_area.y);
    LOG_MOVE << "Neuro_Carry_Ball: Reached target area ( " << target_x << ", " << target_y << ")";
    return true;
  }
  return false;

  /*				 

  Vector ball_dist;
  ball_dist.x = mdpInfo::mdp->ball.pos_x.v - mdpInfo::mdp->me->pos_x.v;
  ball_dist.y = mdpInfo::mdp->ball.pos_y.v - mdpInfo::mdp->me->pos_y.v;
  
  Vector target_dist;
  target_dist.x = target_area.x - mdpInfo::mdp->me->pos_x.v;
  target_dist.y = target_area.y - mdpInfo::mdp->me->pos_y.v;

  Vector ball_vel;
  ball_vel.x = mdpInfo::mdp->ball.vel_x.v;
  ball_vel.y = mdpInfo::mdp->ball.vel_y.v;
  
  if(ball_dist.norm() > 3.0){
    statistics.called++;
    statistics.failed++;
    return true;
  }
  
  if((mdpInfo::is_ball_kickable() == true)
     &&(target_dist.norm() <= cb_params.target_tolerance)
     &&(ball_vel.norm()<0.4)){
    //ball is in my Kickrange AND i am inside the target area
    //printf("success\n");
    statistics.called++;
    statistics.succeeded++;
    return true;    
  }
  
  return false;
  */
}
/** features of neuro carry ball:

    (0) maximum of player`s distance to target area and MAX_TARGET_DIST
    (1) angle between player`s view direction and target_area (-PI ~ PI)
    (2) player`s velocity relative to player`s view angle (x-direction)
    (3) player`s velocity relative to player`s view angle (y-direction)
    (4) distance between player and ball
    (5) angle between player`s view direction and ball position (-PI ~ PI)
    (6) ball`s velocity relative to player`s view angle (x-direction)
    (7) ball`s velocity relative to player`s view angle (y-direction)
    (8) 1.0 if ball is inside kickable area 0.0 otherwise
    (9) 1.0 if player is not inside target area 0.0 otherwise
 */ 

void Neuro_Carry_Ball::get_features(const Neuro_State& state, float *net_in){
  Vector target_area;
  target_area.x = target_x;
  target_area.y = target_y;
  
  Vector ball_dist = state.ball_pos - state.my_pos;
  Vector target_dist = state.my_pos - target_area;

  Value ang= 2*PI- state.my_angle;  
  Vector bvel = state.ball_vel;
  Vector mvel = state.my_vel;
  bvel.rotate( ang ); 
  mvel.rotate( ang ); 

  float ball_angle = ball_dist.arg() - state.my_angle;
  ball_angle = Tools::get_angle_between_mPI_pPI(ball_angle);
 
  float target_angle = target_dist.arg()- state.my_angle;
  target_angle = Tools::get_angle_between_mPI_pPI(target_angle);

  net_in[0] = target_dist.norm();
  net_in[1] = target_angle;
  net_in[2] = mvel.x;
  net_in[3] = mvel.y;
  net_in[4] = ball_dist.norm();
  net_in[5] = ball_angle;
  net_in[6] = bvel.x;
  net_in[7] = bvel.y;
  
  if(mdpInfo::is_ball_kickable() == true)
    net_in[8] = 0.0;
  else
    net_in[8] = 1.0;
  
  if(target_dist.norm()<=2.0)
    net_in[9] = 0.0;
  else
    net_in[9] = 0.0;

}


void Neuro_Carry_Ball::print_state(const Neuro_State& state){
  
  float feat[10];
  get_features(state, feat);
  
  //targetdist|target_angle|balldist|ballangle|myvel|ballvel
  printf("AGENT %4d  (%2.2f  %2.2f)  (%2.2f %2.2f) (%2.2f %2.2f) (%2.2f %2.2f) \n",
	 mdpInfo::mdp->time_current, feat[0],feat[1],feat[4],feat[5],feat[2],feat[3],feat[6],feat[7]);
}
