#include "neuro_dribble2.h"
#include "mdp_info.h"
#include "options.h"
#include "tools.h"
#include "move_factory.h"
#include "logger.h"

#define MAX_TARGET_DIST 8.0
#define LOG_DRIBBLE2_ON

Neuro_Dribble2::Neuro_Dribble2(double p_target_x=0.0, double p_target_y = 0.0){
  target_x = p_target_x;
  target_y = p_target_y;
  
  last_time_executed = -1;
  
  // important: do init first, since net loading might be aborted
  nets[0] = new Net();
  if(nets[0]->load_net("./data/nets_neuro_dribble2/dribble2.net") == FILE_ERROR){
    cout<<"Neuro_Dribble2: No net-file found in directory ./nets - stop loading\n";
    exit(0);
  }
  nets[1] = new Net();
 
  net_counter = 1;

  /* set documentation level */
  params.doku_level = 0;  
  params.costs_failure = 1.0;
  params.costs_success = 0.0;
  
  /* neuro dribble uses dashs */
  params.dash_power_min = 10;
  params.dash_power_inc = 10;
  params.dash_power_max = 100;
  
  /* neuro dribble uses turns */
  params.turn_angle_min = 2*PI/24;
  params.turn_angle_inc = 2*PI/24;
  params.turn_angle_max = 2*PI;

  /* neuro dribble uses kicks */
  params.kick_power_min = 10;
  params.kick_power_inc = 10;
  params.kick_power_max = 100;

  params.kick_angle_min = 0;
  params.kick_angle_inc = 2*PI/36;
  params.kick_angle_max = 2*PI;
 
  params.kick_finetune = 0;
  params.kick_finetune_power = 0;
  params.turn_finetune = 0;

  /* set custom dribble params */
  min_ball_dist = 0.15;
  target_tolerance = 2.0;
  max_target_velocity = 0.4;
   
  reset_statistics();

}

/** destructor method */
Neuro_Dribble2::~Neuro_Dribble2(){
  delete nets[0];
}

/** rule based pre-computing part of the neuro carry ball move  */ 
bool Neuro_Dribble2::decide_rule_based(Cmd & cmd){

  /* compute effective absolute dribble target */
  Vector target = Vector(target_x, target_y) - mdpInfo::my_pos_abs();
  if(target.norm() > 8.0){
    target.normalize(8.0);
    target_x =  mdpInfo::my_pos_abs().x+target.x;
    target_y =  mdpInfo::my_pos_abs().y+target.y;
  }
#ifdef LOG_DRIBBLE2_ON
 LOG_MOVE << "Neuro_Dribble2: Target (Rel)  (" << target.x << ", " << target.y << ")"; 
#endif
  int x1,x2,x3,x4;
  x1=(int)mdpInfo::my_pos_abs().x;
  x2=(int)mdpInfo::my_pos_abs().y;
  x3 = x1 + (int)target.x;
  x4 = x2 + (int)target.y;
  LOG_DRAWLINE<<x1<<" "<<x2<<" "<<x3<<" "<<x4<<" green";
    
  return true;
}


/** specifies the successful terminal states of the neuro carry ball move*/ 
bool Neuro_Dribble2::is_success(const Neuro_State& state){
  if(is_failure(state) == true){
    return false;
  }
  
  /* get abs. target */
  Vector target_area = Vector(target_x, target_y);
  Vector target_dist = state.my_pos-target_area;
  
  /* ball is in my kickrange AND i am inside the target area AND ball vel is small */
  if((state.ball_pos.distance(state.my_pos) < ServerOptions::kickable_area-0.2)
     &&(target_dist.norm() <= 2.0)
     &&(state.ball_vel.norm()<0.2)){
    return true;
  }
  
  return false;
}

/** specifies the unsuccessful terminal states of the neuro carry ball move*/ 
bool Neuro_Dribble2::is_failure(const Neuro_State& state){
  /* get oppponent pos */
  Value dir,dist;
  Vector op_pos,op_vel;
  mdpInfo::opponent_closest_to_me(dist,dir,op_pos); 

  if(dist > 4){
    op_pos = Vector(200,200);
  }
    
  Vector ball_op_dist = op_pos - state.ball_pos;
  Vector ball_dist = state.ball_pos - state.my_pos;
  Vector op_dist =  op_pos - state.my_pos;
  float angle_op_ball_diff = op_dist.arg() - ball_dist.arg();
  angle_op_ball_diff = Tools::get_abs_angle(angle_op_ball_diff);
  
  
  /* if opponent is near => hide ball */
  if((angle_op_ball_diff < 0.6*PI)&&(op_dist.norm()<2.5)){
    return true;
  }
 
  /* dont let ball go outside the kickable area (-buffer) */
  if(ball_dist.norm() > ServerOptions::kickable_area-0.1){
    return true;
  } 
 
  /* dont let ball go inside the opponents kickable area (+buffer) */
  if(ball_op_dist.norm() < ServerOptions::kickable_area+0.1){
    return true;
  }   
 
  /* dont kick ball inside myself (+buffer) */
  if(ball_dist.norm() <  ServerOptions::player_size+0.15){
    return true;
  }   
  
  /* dont bump into opponent */
  if(op_dist.norm() < 2*ServerOptions::player_size+0.1){
    return true;
  }
  
  //if((angle_op_ball_diff < 0.4*PI)&&(ball_op_dist.norm()<KICKABLE_AREA+1.0)){
  //  return true;
  //}
  
  return false;
}


float Neuro_Dribble2::default_V(const Neuro_State& state){
  
  float ball_dist = (state.ball_pos - state.my_pos).norm();
  
  /* dont kick inside myself */
  if(ball_dist < ServerOptions::player_size+0.1){
    return 1.0;
  }
  
  /* ball has left my kick range */
  if(ball_dist > ServerOptions::kickable_area){
    return 1.0;
  }
  
  return 0.5;
}

bool Neuro_Dribble2::final_state(){
   
  Value dir,dist;
  Vector op_pos,op_vel;
  mdpInfo::opponent_closest_to_me(dist,dir,op_pos); 
  if(dist > 4){
    op_pos = Vector(200,200);
  }
 
  if(op_pos.distance(mdpInfo::ball_pos_abs()) < ServerOptions::kickable_area){

#ifdef LOG_DRIBBLE2_ON
    LOG_MOVE << "Neuro_Dribble2: Ball is in opponents kick range!";
#endif
    return true;
  }

  if(!mdpInfo::is_ball_kickable_exact()){
#ifdef LOG_DRIBBLE2_ON
    LOG_MOVE << "Neuro_Dribble2: Lost Ball!";
#endif
    return true;
  }
  
  /* get abs. target */
  Vector target_area = Vector(target_x, target_y);
  Vector my_pos = mdpInfo::my_pos_abs();
  
  if(my_pos.distance(target_area) < 1.5) {
#ifdef LOG_DRIBBLE2_ON
    LOG_MOVE << "Neuro_Dribble2: Reached target area ( " << target_x << ", " << target_y << ")";
#endif 
   return true;
  }
  
  return false;
}



int Neuro_Dribble2::get_opponent_number(const Neuro_State& state){
  int min_number = -1;
  float min_dist = 1000;
  
  for(int i=1;i<11;i++){
    if(!mdpInfo::is_opponent_pos_valid(i)) continue;
    Vector op_pos = mdpInfo::opponent_pos_abs(i);
    float op_ball_dist = state.ball_pos.distance(op_pos);
    if(op_ball_dist < min_dist){
      min_dist = op_ball_dist;
      min_number = i;
    }
  }
  
  return min_number;
}


/** features of neuro_dribble2:

 */  
void Neuro_Dribble2::get_features(const Neuro_State& state, float *net_in){
  /* get abs. target */
  Vector target_area = Vector(target_x, target_y);
  
  Value dir,dist;
  Vector op_pos,op_vel;
  mdpInfo::opponent_closest_to_me(dist,dir,op_pos); 

  
  if(dist > 4){
    op_pos = Vector(200,200);
  }
  
  op_pos = op_pos - target_area;
  Vector ball_vel = state.ball_vel;
  Vector ball_pos = state.ball_pos - target_area;
  Vector my_pos = state.my_pos - target_area;
  Vector my_vel = state.my_vel;
  Value my_view = state.my_angle;
  
  float angle = my_pos.arg();
  /* rotate whole system (pos.y should be zero*/
  my_pos.rotate(2*PI - angle);
  my_vel.rotate(2*PI - angle);
  ball_vel.rotate(2*PI - angle);
  ball_pos.rotate(2*PI - angle);
  op_pos.rotate(2*PI - angle);
  op_vel.rotate(2*PI - angle);
  my_view = Tools::get_angle_between_mPI_pPI((my_view - angle)-PI);
  Vector rel_ball_pos = ball_pos - my_pos;

  if(my_view<0) {
    my_vel.y = -my_vel.y;
    rel_ball_pos.y = -rel_ball_pos.y;
    ball_vel.y = -ball_vel.y;
    op_pos.y = -op_pos.y;
    op_vel.y = -op_vel.y;
  }

  float my_ball_view = rel_ball_pos.arg()-fabs(my_view);
  my_ball_view =  Tools::get_angle_between_mPI_pPI(my_ball_view);

  /* agent features */
  net_in[0] = my_pos.norm(); //distance to target
  net_in[1] = fabs(my_view); //abs. relative view angle to target
  net_in[2] = my_vel.x; //velocity in x direction
  net_in[3] = my_vel.y;//velocity in y direction
  
  /* ball features */
  net_in[4] = rel_ball_pos.norm();
  net_in[5] = my_ball_view;
  net_in[6] = ball_vel.x;
  net_in[7] = ball_vel.y;
  
  /* op features */
  Vector rel_op_pos = my_pos-op_pos;
  if(rel_op_pos.norm() < 4.0) {
    float ang_diff = rel_ball_pos.arg()-rel_op_pos.arg();
    ang_diff =  Tools::get_angle_between_mPI_pPI(ang_diff);
    net_in[8] = rel_op_pos.x;//PI-fabs(ang_diff);
    net_in[9] = rel_op_pos.y;//5.0-(rel_ball_pos-rel_op_pos).norm();
  }
  else {
    net_in[8] = -4.0;
    net_in[9] = 4.0;
  }
  /* additional features */
  if(rel_ball_pos.norm() < ServerOptions::kickable_area)
    net_in[10] = 1.0;
  else
    net_in[10] = 0;
}

/** generate action set */
void Neuro_Dribble2::generate_actions(const Neuro_State& state){
  action_counter = 0;
  double power, angle;

  /* add turn commands to action set */
  for(angle=params.turn_angle_min; angle<params.turn_angle_max; angle+=params.turn_angle_inc){
    actions[action_counter].unset_lock();
    actions[action_counter].unset_cmd();
    
    actions[action_counter].set_turn(angle);
    action_counter++;
  }

  /* add dash  commands to action set */
  for(power=params.dash_power_min; power<=params.dash_power_max; power+=params.dash_power_inc){    
    actions[action_counter].unset_lock();
    actions[action_counter].unset_cmd();
    
    actions[action_counter].set_dash(power);
    action_counter++;
  }
 
  /* add kick commands to action set */
  if(mdpInfo::is_ball_kickable_exact() == true){
    
    for(power=params.kick_power_min; power<=params.kick_power_max; power+=params.kick_power_inc){
      for(angle=params.kick_angle_min; angle<params.kick_angle_max; angle+=params.kick_angle_inc){
	actions[action_counter].unset_lock();
	actions[action_counter].unset_cmd();
	
	actions[action_counter].set_kick(power, angle);
	action_counter++;
      }
    }
  }
}

void Neuro_Dribble2::print_state(const Neuro_State& state){
  
  float feat[10];
  get_features(state, feat);
  
  //targetdist|target_angle|balldist|ballangle|myvel|ballvel
  printf("AGENT %4d  (%2.2f  %2.2f)  (%2.2f %2.2f) (%2.2f %2.2f) (%2.2f %2.2f) \n",
	 mdpInfo::mdp->time_current, feat[0],feat[1],feat[4],feat[5],feat[2],feat[3],feat[6],feat[7]);
}
