#include "neuro_go2pos.h"
#include "mdp_info.h"
#include "options.h"
#include "tools.h"
#include "move_factory.h"
#include "logger.h"

#define MAX_GO2POS_DIST 5.0

//#define LOG_GO2POS_ON

Neuro_Go2Pos::Neuro_Go2Pos(float p_target_x, float p_target_y, float p_max_dash_power = 100, float p_min_dash_power=40, float p_inc_dash_power = 20 ){
  /* take params */
  target_x = p_target_x;
  target_y = p_target_y;
  max_dash_power = p_max_dash_power;
  min_dash_power = p_min_dash_power;
  inc_dash_power = p_inc_dash_power;

  /* init some other stuff */
  last_time_executed = -1;

  /* init statistics */
  stat_extend.intercept_slow =  stat_extend.leave_offside =  
  stat_extend.offside_trap = stat_extend.block_ballholder = stat_extend.defend = 
  stat_extend.offer_for_pass = stat_extend.attack = stat_extend.no_spec=0;
  
  /* load neural network */
  char *netname = "./data/nets_neuro_go2pos/go2pos.net";
  
  // important: do init first, since net loading might be aborted
  nets[0] = new Net();
  if(nets[0]->load_net(netname) == FILE_ERROR){
    cout<<"Neuro_Go2Pos: No net-file found in directory ./nets - stop loading\n";
    return;
  }
  
  /* set params */
  params.doku_level = 0;  
  params.costs_failure = 1.0;
  params.costs_success = 0.0;
   
  /* neuro_go2pos uses dashs */
  params.dash_power_min = 40;
  params.dash_power_inc = 20;
  params.dash_power_max = 100;
  
  /* neuro_go2pos uses turns */
  params.turn_angle_min = 2*PI/16;
  params.turn_angle_inc = 2*PI/16;
  params.turn_angle_max = 2*PI;

  /* neuro_go2pos uses no kicks */
  params.kick_power_min = 0;
  params.kick_power_inc = 0;
  params.kick_power_max = 0;

  params.kick_angle_min = 0;
  params.kick_angle_inc = 0;
  params.kick_angle_max = 0;

  params.kick_finetune = 0;
  params.kick_finetune_power = 0;
  params.turn_finetune = 0;

  /* set csutom go2pos params */
  target_tolerance = 1.0;
}

/** destructor method */
Neuro_Go2Pos::~Neuro_Go2Pos(){
  delete nets[0];
}

bool Neuro_Go2Pos:: decide_rule_based(Cmd & cmd){

  //#ifdef LOG_GO2POS_ON
  LOG_MOVE << "Neuro_Go2Pos: Go to position (" << target_x << " , " << target_y << ")";
  LOG_DRAWCIRCLE<<(int) target_x<<" "<<(int) target_y<<" 1 green";
  //#endif
  /* Check if this is the first step of the move */
  if(last_time_executed +1 != mdpInfo::mdp->time_current){
    last_time_executed = mdpInfo::mdp->time_current;
  }
  
  /* check if target inside learned horizon */
  Vector target_area = Vector(target_x, target_y);
  float target_dist = mdpInfo::my_distance_to(target_area);
  if(target_dist > MAX_GO2POS_DIST){
    cutoff_value = target_dist - MAX_GO2POS_DIST;
  }
  else {
    cutoff_value = 0;
  }
  
  /* which opponent player is the worst obstacle? */
  Vector temp_target = target_area;
  if(cutoff_value > 0){
    Vector temp = target_area - current_state.my_pos;
    temp.normalize(MAX_GO2POS_DIST);
    temp_target = current_state.my_pos + temp;
  }
  
  Vector my_target_dist = temp_target - current_state.my_pos;
  float min_alpha = 2.0;
  int opponent = 0;
  for(int i=1; i<=11; i++){
    
    if(mdpInfo::is_opponent_pos_valid(i)){
      Vector op_pos = mdpInfo::opponent_pos_abs(i);
      Vector my_op_dist = op_pos - current_state.my_pos;
      /* compute distance of obstacle to the line me/target */
      float dist = my_target_dist.norm()*my_target_dist.norm();
      float alpha = ((my_target_dist.x*my_op_dist.x)+(my_target_dist.y*my_op_dist.y))/dist;
      /* only consider opponents near the line me/target */
#ifdef LOG_GO2POS_ON
      LOG_MOVE_D << "Neuro_Go2Pos: Checking player " << i << " alpha: " << alpha;
#endif
      if((alpha>=0.0)&&(alpha<1.0)) {
	Vector cdist;
	cdist.x = alpha*my_target_dist.x - my_op_dist.x;
	cdist.y = alpha*my_target_dist.y - my_op_dist.y;
	if(cdist.norm() < 2.0){
	  if(alpha < min_alpha){
	    min_alpha = alpha;
	    opponent = i;
	  }
	}
      }
    }
  }
  
  if(opponent > 0){
    current_state.opnt1_pos = mdpInfo::opponent_pos_abs(opponent);
#ifdef LOG_GO2POS_ON
    LOG_MOVE_D << "Neuro_Go2Pos: Player " << opponent << " detected as an obstacle.";
#endif
  }
  else {
    current_state.opnt1_pos = Vector(60,60);
  }

  return true;
}

void Neuro_Go2Pos::generate_actions(const Neuro_State& state){
  action_counter = 0;
  double power, angle;

  /* add turn commands to action set */
  Vector ball_dist = state.ball_pos - state.my_pos;
  float ball_view_angle = ball_dist.arg() - state.my_angle;
  for(angle=params.turn_angle_min; angle<params.turn_angle_max; angle+=params.turn_angle_inc){
    actions[action_counter].unset_lock();
    actions[action_counter].unset_cmd();

    float turn_angle = ball_view_angle+angle;
    turn_angle = Tools::get_angle_between_null_2PI(turn_angle);    
    actions[action_counter].set_turn(turn_angle);
    action_counter++;
  }

  /* add dash  commands to action set */
  
  for(power=min_dash_power; power<=max_dash_power; power+=inc_dash_power){    
    actions[action_counter].unset_lock();
    actions[action_counter].unset_cmd();
    
    actions[action_counter].set_dash(power);
    action_counter++;
  }

}

/** specifies the successful terminal states of the neuro_go2pos move */ 
bool Neuro_Go2Pos::is_success(const Neuro_State& state){
  Vector target_area = Vector(target_x, target_y);
  Vector target_dist = state.my_pos-target_area;
  
  if(target_dist.norm() <= target_tolerance){
    /* reached the target area */
    return true;
  }

  return false;
}

/** specifies the unsuccessful terminal states of the neuro_go2pos move */
bool Neuro_Go2Pos::is_failure(const Neuro_State& state){
  
  Vector pos;
  /* ridi: replaced that by new access which considers players in feelrange
  int opponent = mdpInfo::opponent_closest_to_me();
  if(opponent > 0){
    if(state.my_pos.distance(mdpInfo::opponent_pos_abs(opponent)) < 0.3){
      return true;
    }
  }
 return false;
  */
  mdpInfo::opponent_closest_to_me(pos);
  if(state.my_pos.distance(pos) < 0.3)
    return true;
  return false;
}

/** features of neuro_go2pos */ 
void Neuro_Go2Pos::get_features(const Neuro_State& state, float *net_in){
  Vector target_area = Vector(target_x, target_y);
  Vector temp_target = target_area;
  if(cutoff_value > 0){
    Vector temp = target_area - state.my_pos;
    temp.normalize(MAX_GO2POS_DIST);
    temp_target = state.my_pos + temp;
  }
  
  Vector my_op_dist = state.opnt1_pos - state.my_pos;
  Vector my_target_dist = temp_target - state.my_pos;
  Vector my_vel = state.my_vel;
  Value my_angle = state.my_angle;
  
  float my_target_angle = my_target_dist.arg() - my_angle;
  my_target_angle = Tools::get_angle_between_mPI_pPI(my_target_angle);
  my_vel.rotate(2*PI - my_target_dist.arg());
  
  net_in[0] = my_target_dist.norm();
  net_in[1] = my_target_angle;
  net_in[2] = my_vel.x;
  net_in[3] = my_vel.y;
  
  /* compute distance of obstacle to the line me/target */
  float dist = my_target_dist.norm()*my_target_dist.norm();
  float alpha = ((my_target_dist.x*my_op_dist.x)+(my_target_dist.y*my_op_dist.y))/dist;
  
  /* default: ignore obstacles */
  net_in[4] = 3.0;
  net_in[5] = 3.0;
  net_in[6] = 0;
  net_in[7] = PI/2;    
  
  if((alpha>=0.0)&&(alpha<1.0)){
    Vector cdist;
    cdist.x = alpha*my_target_dist.x - my_op_dist.x;
    cdist.y = alpha*my_target_dist.y - my_op_dist.y; 
    float my_op_view_angle = my_op_dist.arg() - my_angle;
    my_op_view_angle = Tools::get_angle_between_mPI_pPI(my_op_view_angle);
    
    float angle_diff1 = my_target_dist.arg() - my_op_dist.arg();
    angle_diff1 = Tools::get_angle_between_mPI_pPI(angle_diff1);
    
    net_in[4] = my_op_dist.norm();
    net_in[5] = cdist.norm();  
    net_in[6] = 0;//my_op_view_angle;
    net_in[7] = angle_diff1;
  }
}

/** chrarcter representation of state */
void Neuro_Go2Pos::print_state(const Neuro_State& state){

  float features[6];
  get_features(state, features);
  
  printf("Attack step: %d:  ball_dist %.2f  ball_angle %.2f  ball_vel (%.2f %.2f) my_vel (%.2f %.2f) \n",
	 mdpInfo::mdp->time_current, features[0],  features[1],  features[4],  features[5], features[2],  features[3]);
}

/** final state of neuro_go2pos */
bool Neuro_Go2Pos::final_state(){
  Vector target_area = Vector(target_x, target_y);
  if(mdpInfo::my_distance_to(target_area) <  target_tolerance){
#ifdef LOG_GO2POS_ON
    LOG_MOVE << "Neuro_Go2Pos: Reached the target area.";
#endif
    last_time_executed = -1;
    return true;
  }
  
  if(mdpInfo::is_ball_kickable_exact()){
#ifdef LOG_GO2POS_ON
    LOG_MOVE << "Neuro_Go2Pos: Ball is inside kick range. terminate go2pos move.";
#endif
    last_time_executed = -1;
    return true;
  }
  
  return false;           
}

void Neuro_Go2Pos::execute(Cmd & cmd){
    //fill statistics
  switch(mdpInfo::get_my_intention()){
  
  case DECISION_TYPE_INTERCEPTSLOWBALL: 
    stat_extend.intercept_slow++;
    break;
  case DECISION_TYPE_OFFSIDETRAP:
    stat_extend.offside_trap++;
    break;
  case DECISION_TYPE_LEAVE_OFFSIDE:
    stat_extend.leave_offside++;
    break;
  case DECISION_TYPE_DEFEND:
    stat_extend.defend++;
    break;
  case DECISION_TYPE_BLOCKBALLHOLDER:
    stat_extend.block_ballholder++;
    break;
  case DECISION_TYPE_ATTACK:
    stat_extend.attack++;
    break;
  case DECISION_TYPE_OFFER_FOR_PASS:
    stat_extend.offer_for_pass++;
    break;
  default:
    stat_extend.no_spec++;
    break;
  }

  Neuro_Move::execute(cmd);
}

void Neuro_Go2Pos::print_statistics(){
  cout << "\nMove: Neuro-Go2Pos";
  cout << "\n  Executed Total    : " << statistics.called;
  cout << "\n  Average steps/call: " << statistics.called/(float)statistics.succeeded;
  cout << "\n  INTERCEP_SLOWBALL : " << stat_extend.intercept_slow;
  cout << "\n  LEAVE_OFFSIDE     : " << stat_extend.leave_offside;
  cout << "\n  PULL_OFFSIDE_TRAP : " << stat_extend.offside_trap;
  cout << "\n  BLOCK_BALLHOLDER  : " << stat_extend.block_ballholder;
  cout << "\n  DEFEND            : " << stat_extend.defend;
  cout << "\n  ATTACK            : " << stat_extend.attack;
  cout << "\n  OFFER_FOR_PASS    : " << stat_extend.offer_for_pass;
  cout << "\n  not specified     : " << stat_extend.no_spec;
}
