#include "neuro_goalie_intercept.h"
#include "mdp_info.h"
#include "options.h"
#include "tools.h"
#include "move_factory.h"
#include "logger.h"

/** neuro_intercept_ball constructor method */
Neuro_Goalie_Intercept::Neuro_Goalie_Intercept(){
  
  // important: do init first, since net loading might be aborted
  nets[0] = new Net();

  if(nets[0]->load_net("./data/nets_neuro_intercept/goalie_intercept.net") == FILE_ERROR){
    cout<<"NeuroGoalieInterceptBall: No net-file found in directory ./nets - stop loading\n";
    exit(0);
  }

  net_counter = 1;
  
   
  /* set params */
  params.doku_level = 0;  
  params.costs_failure = 1.0;
  params.costs_success = 0.0;
   
  /* neuro intercept uses dashs */
  params.dash_power_min = 100;
  params.dash_power_inc = 10;
  params.dash_power_max = 100;
  
  /* neuro intercept uses turns */
  params.turn_angle_min = 2*PI/18;
  params.turn_angle_inc = 2*PI/18;
  params.turn_angle_max = 2*PI;

  /* neuro intercept uses no kicks */
  params.kick_power_min = 0;
  params.kick_power_inc = 0;
  params.kick_power_max = 0;

  params.kick_angle_min = 0;
  params.kick_angle_inc = 0;
  params.kick_angle_max = 0;

  params.kick_finetune = 0;
  params.kick_finetune_power = 0;
  params.turn_finetune = 2*PI/(18*18);

  /* set custom intercept_ball params */
  catchable_tolerance = 0.05;
}

/** destructor method */
Neuro_Goalie_Intercept::~Neuro_Goalie_Intercept(){
  delete nets[0];
}

/** specifies the successful terminal states of the neuro intercept ball move
    intercept ball is finished successful if the ball is inside the 
    players kick range.
*/ 
bool Neuro_Goalie_Intercept::is_success(const Neuro_State& state){
  
  Vector ball_dist = state.my_pos - state.ball_pos;
  if(ball_dist.norm() <= ServerOptions::catchable_area_l-catchable_tolerance){
    //ball is in my kick range
    return true;
  }
  
  return false;
}

/** specifies the unsuccessful terminal states of the neuro intercept ball move
    there are no unsuccessful states for intercept ball
*/ 
bool Neuro_Goalie_Intercept::is_failure(const Neuro_State& state){
  return false;
}
void Neuro_Goalie_Intercept::generate_actions(const Neuro_State& state){
  action_counter = 0;

  /* add turn commands to action set */
 
  generate_turns(state, 0, params.turn_angle_max, params.turn_angle_inc);
 
  /* add dash  commands to action set */
  generate_dashes(state, params.dash_power_min, params.dash_power_max, params.dash_power_inc);

}

/** features of neuro intercept ball:
    (0) distance between player and ball
    (1) angle between player`s view direction and ball (-PI ~ PI)
    (2) player velocity x-direction (rotated)
    (3) player velocity y-direction (rotated)
    (4) ball velocity x-direction (rotated)
    (5) ball velocity y-direction (rotated)
 */ 
void Neuro_Goalie_Intercept::get_features(const Neuro_State& state, float *net_in){
  
  Vector ball_dist = state.ball_pos - state.my_pos;
  Vector ball_vel = state.ball_vel;
  Vector my_vel = state.my_vel;

  float ball_angle = ball_dist.arg();
  float ball_view_angle = ball_dist.arg() - state.my_angle;  

  ball_vel.rotate(2*PI-ball_angle);
  my_vel.rotate(2*PI-ball_angle);


  ball_view_angle = Tools::get_angle_between_mPI_pPI(ball_view_angle);
  
  net_in[0] = ball_dist.norm();
  net_in[1] = ball_view_angle;
  
  net_in[2] = my_vel.x;
  net_in[3] = my_vel.y;
  
  net_in[4] = ball_vel.x;
  net_in[5] = ball_vel.y;

}

/** specifies termination conditions of neuro_intercept_ball */
bool Neuro_Goalie_Intercept::final_state(){
  //int my_number = mdpInfo::mdp->me->number;
  
  /* terminate intercept if a teammate (or me) is able to kick the ball */
  //int teammate =  mdpInfo::teammate_closest_to_ball();
 
  if(mdpInfo::is_ball_catchable()){
    LOG_MOVE << "NeuroGoalieIntercept: Ball is catchable.";
    return true;
  }
  
  return false;           
}

void Neuro_Goalie_Intercept::print_state(const Neuro_State& state){
  
  float features[6];
  get_features(state, features);
  
  printf("Intercept step: %d:  ball_dist %.2f  ball_angle %.2f  ball_vel (%.2f %.2f) my_vel (%.2f %.2f) \n",
	 mdpInfo::mdp->time_current, features[0],  features[1],  features[4],  features[5], features[2],  features[3]);
}

