#include "neuro_hold_ball.h"
#include "mdp_info.h"
#include "options.h"
#include "tools.h"
#include "move_factory.h"
#include "logger.h"

#define LOG_INFO2(X)
//#define LOG_INFO2 LOG_INFO

/** constructor method */
Neuro_Hold_Ball::Neuro_Hold_Ball(){
 
  char *netname = "./data/nets_neuro_hold_ball/holdball.net";
  
  // important: do init first, since net loading might be aborted
  nets[0] = new Net();
  if(nets[0]->load_net(netname) == FILE_ERROR){
    cout<<"Neuro_Hold_Ball: No net-file found in directory ./nets - stop loading\n";
    return;
  }
  
  /* set neuro_stop_ball params */
  params.costs_failure = 1.0;
  params.costs_success = 0.0;
  params.doku_level = 0;  // 0 = silence
  
  /* neuro_stop_ball uses no dashs */
  params.dash_power_min = 0;
  params.dash_power_inc = 0;
  params.dash_power_max = 0;
  
  /* neuro_stop_ball uses no turns */
  params.turn_angle_min = 0;
  params.turn_angle_inc = 0;
  params.turn_angle_max = 0;

  /* neuro_stop_ball kicks */
  params.kick_power_min = 10;
  params.kick_power_inc = 10;
  params.kick_power_max = 100;
  params.kick_angle_min = 0;
  params.kick_angle_inc = 2*PI/45;
  params.kick_angle_max = 2*PI;

 /* set hold ball params to defaults */
  ball_safe_distance = 0.15;
  kickable_safe_distance = 0.15;  
  success_velocity = 0.05;
  success_angle_diff = 0.9*PI;
}

/** destructor method */
Neuro_Hold_Ball::~Neuro_Hold_Ball(){
  delete nets[0];
}
/** rule based pre-computing part */ 
bool Neuro_Hold_Ball::decide_rule_based(Cmd & cmd){
  /* determine nearest opponent */
  op_position = opponent_pos(current_state);
  Vector rel_op_pos = op_position-current_state.my_pos;
  float op_dist = rel_op_pos.norm();
  float op_angle = rel_op_pos.arg();   

  /*
  if((holdball_mode == TURN_MODE)&&(op_dist>2.5)){
    op_position = current_state.my_pos + Vector(2,0);
    success_angle_diff = 0*PI;
    success_velocity = 0.1;
  }
  else if((holdball_mode == TURN_MODE)&&(op_dist>2.0)){
    op_position = current_state.my_pos + Vector(2,0);
    success_angle_diff = 0.6*PI;
    success_velocity = 0.1;
  }
  */
  // ridi: the above seems to be too lax, take distance to the ball to feel safe!
  float op_dist2ball = (op_position-current_state.ball_pos).norm();
  Vector tmppos;
  if((mdpInfo::opponent_closest_to_me(tmppos) == mdpInfo::opponent_goalie_number()) ||
     (mdpInfo::opponent_closest_to_me(tmppos)<0 && 
      mdpInfo::is_object_in_their_penalty_area(mdpInfo::my_pos_abs())))
    op_dist2ball -= 1.5; // in case the opponent is a goalie, be more cautious
  if((holdball_mode == TURN_MODE)&&(op_dist2ball>2.5)){
    op_position = current_state.my_pos + Vector(2,0);
    success_angle_diff = 0*PI;
    success_velocity = 0.1;
  }
  else if((holdball_mode == TURN_MODE)&&(op_dist2ball>2.0)){
    op_position = current_state.my_pos + Vector(2,0);
    success_angle_diff = 0.6*PI;
    success_velocity = 0.1;
  }
  else{
    // use default values:
    success_angle_diff = 0.9*PI;
    success_velocity = 0.05;    
  }



  if(holdball_mode == HOLD_MODE){
    success_angle_diff = 0.9*PI;
    success_velocity = 0.05;
  }
  
  rel_op_pos = op_position-current_state.my_pos;
  op_dist  = rel_op_pos.norm();
  op_angle = rel_op_pos.arg();   

  LOG_MOVE << "Hold_Ball in mode "<< holdball_mode << " anglediff+ " << success_angle_diff <<"  Nearest Opponent:  Dist "<< op_dist<< "  Angle " << op_angle;

  LOG_INFO2("Hold_Ball in mode "<< holdball_mode << " anglediff+ " << RAD2DEG(success_angle_diff)
	    <<"  Nearest Opponent:  Dist "<< op_dist<< "  Angle " <<RAD2DEG(op_angle)
	    <<" dist2 ball "<<op_dist2ball
	    <<" rel_op_pos "<<rel_op_pos);

  Cmd_Main command;
  command.unset_lock();
  command.unset_cmd();
  command.set_turn(90);
  Neuro_State p_state;
  
  model(current_state, command, p_state);
  if(!is_failure(p_state)&&p_state.ball_vel.norm()<0.7){
    LOG_MOVE << "HoldBall_Ball: One Step Away From Success"; 
   
    float angle = turn_to_angle - current_state.my_angle;
    angle = Tools::get_angle_between_mPI_pPI(angle);
    if(fabs(angle) < PI/16) return true;
    angle = angle * (1.0 + (ServerOptions::inertia_moment * (current_state.my_vel.norm())));
 
    if (angle > PI) angle = PI;
    if (angle < -PI) angle = -PI;

    angle = Tools::get_angle_between_null_2PI(angle);
    cmd.cmd_main.set_turn(angle);
    return false;
  }

  return true;
}

void Neuro_Hold_Ball::generate_actions(const Neuro_State& state){
  float angle,power;
  action_counter = 0;
  
  /* add kick commands to action set */
  for(power=params.kick_power_min; power<=params.kick_power_max; power+=params.kick_power_inc){
    for(angle=params.kick_angle_min; angle<params.kick_angle_max; angle+=params.kick_angle_inc){
      actions[action_counter].unset_lock();
      actions[action_counter].unset_cmd();
      
      actions[action_counter].set_kick(power, angle);
      action_counter++;
    }
  }

  for(power=1; power<=7; power+=1){
    for(angle=params.kick_angle_min; angle<params.kick_angle_max; angle+=params.kick_angle_inc){
      actions[action_counter].unset_lock();
      actions[action_counter].unset_cmd();
      
      actions[action_counter].set_kick(power, angle);
      action_counter++;
    }
  }
 
}

Vector Neuro_Hold_Ball::opponent_pos(const Neuro_State& state){
  /*
  int op_idx = mdpInfo::opponent_closest_to(state.ball_pos);
  Vector op_pos = Vector(-52.5,0);
  if(op_idx > 0){
    op_pos = mdpInfo::opponent_pos_abs(op_idx); 
  }
  
  Vector rel_op_pos = op_pos-state.my_pos; 
  if(rel_op_pos.norm() > 3.0){
    rel_op_pos.normalize(3.0);
    op_pos = state.my_pos + rel_op_pos;
  }
  */
  Vector op_pos;
  mdpInfo::opponent_closest_to_me(op_pos);
  Vector rel_op_pos = op_pos-state.my_pos; 
  if(rel_op_pos.norm() > 3.0){
    rel_op_pos.normalize(3.0);
    op_pos = state.my_pos + rel_op_pos;
  }
  return op_pos;
}

/** specifies the successful terminal states of the neuro stop ball move

    hold ball is finished successful if the ball is inside the 
    players kick range and both, player and ball are slow enough
*/ 
bool Neuro_Hold_Ball::is_success(const Neuro_State& state){
  
  if(is_failure(state) == true) return false;

  Vector op_pos = opponent_pos(state);
  Vector op_dist = op_pos - state.my_pos;
  Vector ball_dist = state.ball_pos - state.my_pos; 
  
  float angle_op_ball_diff = op_dist.arg() - ball_dist.arg();
  angle_op_ball_diff = Tools::get_abs_angle(angle_op_ball_diff);
    
  if((angle_op_ball_diff > success_angle_diff)&&
     (state.ball_vel.norm() < success_velocity)){
    return true;
  }
  
  return false;
}
bool Neuro_Hold_Ball::is_optimum(const Neuro_State& state, float V, float Vmin){
  if(V == Vmin){
    float goal_direction = (Vector(52.5,0) - state.my_pos).arg();
    float goal_view = Tools::get_abs_angle(goal_direction-state.my_angle);
    if(goal_view > PI/16){
      return false;
    }
    if(state.ball_vel.norm() < best_vel){
      best_vel = state.ball_vel.norm();
      return true;
    }
    return false;
  }
  else if(V < Vmin){
    return true;
  }
  return false;
}

/** specifies the unsuccessful terminal states of the neuro stop ball move
    
 */ 
bool Neuro_Hold_Ball::is_failure(const Neuro_State& state){
  
  Vector ball_dist = state.ball_pos - state.my_pos;
  if(ball_dist.norm() > ServerOptions::kickable_area-kickable_safe_distance){
    return true;
  }
  
  if(ball_dist.norm() <  ServerOptions::player_size+ball_safe_distance){
    return true;
  }
  
 
  Vector op_pos = op_position;
  Vector op_ball_dist = state.ball_pos - op_pos;
  if(op_ball_dist.norm() < ServerOptions::kickable_area+0.05){
    return true;
  }

  /*
  for(int i=1;i<=11;i++){
    if(!mdpInfo::is_opponent_pos_valid(i)) continue;
    Vector op_pos = mdpInfo::opponent_pos_abs(i);
    if(op_pos.distance(state.ball_pos) > ServerOptions::kickable_area) continue;
    return true;
  }
  */


  return false;
}
/** features of neuro stop ball:
    
    (0) relative ball position x (rotated)
    (1) relative ball position y (rotated)
    (2) player velocity x-direction (rotated)
    (3) player velocity y-direction (rotated)
    (4) ball velocity x-direction (rotated)
    (5) ball velocity y-direction (rotated)
*/ 
void Neuro_Hold_Ball::get_features(const Neuro_State& state, float *net_in){
  Vector op_pos = op_position;
  /* relative features version */
  Vector ball_vel = state.ball_vel;
  Vector my_pos = state.my_pos - op_pos;
  Vector ball_pos = state.ball_pos - op_pos;
  Vector my_vel = state.my_vel;
  Value my_view = state.my_angle;
  
  float angle = my_pos.arg();
  /* rotate whole system (pos.y should be zero*/
  my_pos.rotate(2*PI - angle);
  ball_pos.rotate(2*PI - angle);
  my_vel.rotate(2*PI - angle);
  ball_vel.rotate(2*PI - angle);
  my_view = Tools::get_angle_between_mPI_pPI((my_view - angle)-PI);
  
  Vector rel_ball_pos = ball_pos-my_pos;
  
  if(my_view<0){
    my_vel.y = -my_vel.y;
    ball_vel.y = -ball_vel.y;
    rel_ball_pos.y = -rel_ball_pos.y;
  }
  
  net_in[0] = my_pos.norm(); //distance to target
  net_in[1] = fabs(my_view); //abs. relative view angle to target
  
  net_in[2] = ball_vel.x; 
  net_in[3] = ball_vel.y;
  
  net_in[4] = rel_ball_pos.x; 
  net_in[5] = rel_ball_pos.y;
  
    /*
      Vector rel_op_pos = op_pos-state.my_pos;
      Vector rel_ball_pos = state.ball_pos-state.my_pos;
      Vector ball_vel = state.ball_vel;
      float angle_diff = rel_ball_pos.arg()-rel_op_pos.arg();
      angle_diff = fabs(Tools::get_angle_between_mPI_pPI(angle_diff));
      float my_view_angle = state.my_angle;
      ball_vel.rotate(2*PI-my_view_angle);
      rel_ball_pos.rotate(2*PI-my_view_angle);
      rel_op_pos.rotate(2*PI-my_view_angle);
      //Vector op_vel = state.op_team[0].vel;
      //op_vel.rotate(2*PI-my_view_angle);
      
      net_in[0] = rel_ball_pos.x;
      net_in[1] = rel_ball_pos.y;
      net_in[2] = ball_vel.x;
      net_in[3] = ball_vel.y;
  
      net_in[4] = rel_op_pos.x;
      net_in[5] = rel_op_pos.y;
      net_in[6] = angle_diff;
      net_in[7] = (rel_op_pos-rel_ball_pos).norm();
    */
}


void Neuro_Hold_Ball::print_state(const Neuro_State& state){

  float features[6];
  get_features(state, features);
  
  printf("StopBall step: %d:  ball_dist %.2f  ball_angle %.2f  ball_vel (%.2f %.2f) \n",
	 mdpInfo::mdp->time_current, features[0],  features[1],  features[4],  features[5]);
}

bool Neuro_Hold_Ball::final_state(){
 
  if(mdpInfo::is_ball_kickable_exact() == false){
    LOG_MOVE << "Neuro_Hold_Ball: ball has left the kickrange!";
    return true;
  }
   
  //int op_idx = mdpInfo::opponent_closest_to_ball();
  

  return false;
}
  
void Neuro_Hold_Ball::check_decision(Cmd & cmd){
  // ridi: just to log
  Value angle,power;
  switch( cmd.cmd_main.get_type()) {
  case Cmd_Main::TYPE_KICK:
    cmd.cmd_main.get_kick(power,angle);
    //LOG_INFO2("Hold ball decided to kick "<<power<<" in dir "<<RAD2DEG(angle));
    break;
  case Cmd_Main::TYPE_TURN:
    cmd.cmd_main.get_turn(angle);
    //LOG_INFO2("Hold ball decided to turn "<<RAD2DEG(angle));
    break;
  default:
    //LOG_INFO2("Hold ball strange: no kick an no turn selected ");
    break;
  }
  
}
