#include "neuro_kick.h"
#include "mdp_info.h"
#include "options.h"
#include "move_factory.h"
#include "logger.h"
#include "tools.h"

#define RESOLUTION_DEG 20

#define TOLERANCE_DIRECTION 0.05 // specifies max. allowed deviation from target direction
#define TOLERANCE_VELOCITY 0.2 // specifies max. allowed deviation from target velocity

#define EPSILON_1 0.05  // balldistance must be smaller than KICKABLE_AREA - epsilon 1
#define EPSILON_2 0.1  // balldistance must be larger than PLAYER_SIZE + epsilon 2

#define POWER_INC (100./4.)
#define ANGLE_INC (2*PI/60.)

//#define LOG_KICK_ON


Neuro_Kick::Neuro_Kick(int mode = 0, Value p1 =2.0, Value p2 = 0.0, Value p3 = 0.0){ 
  int i,j;
  char netname[1000];
  
  // important: do init first, since net loading might be aborted
  last_time_executed = -1;
  new_sequence_started = true;

  Vector diff;
  switch(mode){
  case 0:  // shoot relativ to player's view angle
    target_velocity = p1;
    target_direction = p2;
    break;
  case 1: // shoot to absolute position with initial velocity p1
    diff.x = p2;
    diff.y = p3;
    //termine angle between my pos and targetpos
    diff = diff - mdpInfo::my_pos_abs();
    target_direction = diff.arg();
    target_velocity = p1;
    break;
  case 2: // shoot to absolute position with final velocity (v at final pos) p1
    diff.x = p2;
    diff.y = p3;
    // determine angle between my pos and targetpos
    diff = diff - mdpInfo::my_pos_abs();
    target_direction = diff.arg();
    // determine initial kick velocity  
    target_velocity = (1-ServerOptions::ball_decay) * diff.norm() 
      + ServerOptions::ball_decay * p1;
    break;
  case 3: //shoot 2 goal (p2 == left, middle or right)
    diff = ServerOptions::their_goal_pos;
    if(p2 == 0) // left corner
      diff.y += (ServerOptions::goal_width /2. -1);
    else if(p2 == 2) // right corner
      diff.y -= (ServerOptions::goal_width /2. -1);
    else
      diff.y = 0; // middle of goal
    // determine angle between my pos and targetpos
    diff = diff - mdpInfo::my_pos_abs();
    target_direction = diff.arg();
    target_velocity = 2.0; 
    break;
  default:
    cerr <<"Neuro Kick ERROR: Mode "<<mode<<" not implemented"<<endl;
    break;
  }
    
    

  net_counter = 0;
  action_counter = 0;
  for(j=0;j<3;j++){ // three different initial speeds: normal, fast, slow
    for(i=0;i<360;i+=RESOLUTION_DEG){
      //cout<<"Neuro Kick: Loading nets\n";
      nets[net_counter] = new Net();
      if(j==0) // nets for speed 1<= speed <= kick_params.high_speed
	sprintf(netname,"./data/nets_neuro_kick/nk_%d_2.net",i);
      if(j==1) // nets for max. speed > params.high_speed
	sprintf(netname,"./data/nets_neuro_kick/nk_%d_2.5.net",i);
      if(j==2) // nets for slow speed <1.0
	sprintf(netname,"./data/nets_neuro_kick/nk_%d_0.7.net",i);
      if(nets[net_counter]->load_net(netname) == FILE_ERROR){
	cout<<"NeuroKick: No net-files found in directory ./nets - stop loading\n";
	return;
      }
      net_counter ++;
      if(net_counter>=1000){
	cout<<"NeuroKick: too many nets (>1000)\n";
	return;
      }
    }    
  } // end load nets
  
  
  /* set neuro kick params */
  params.costs_failure = 1.0;
  params.costs_success = 0.0;
  params.doku_level = 0;  // 0 = silence
  
  /* neuro kick uses no dashs */
  params.dash_power_min = 0;
  params.dash_power_inc = 0;
  params.dash_power_max = 0;
  
  /* neuro kick uses no turns */
  params.turn_angle_min = 0;
  params.turn_angle_inc = 0;
  params.turn_angle_max = 0;

  /* neuro kick uses kicks */
  params.kick_power_min = POWER_INC;
  params.kick_power_inc = POWER_INC;
  params.kick_power_max = 100;
  params.kick_angle_min = 0;
  params.kick_angle_inc = ANGLE_INC;
  params.kick_angle_max = 2*PI;
 
  params.kick_finetune = 1/180. * PI;  // 0: finetune off. if >0: angle inc in RAD
  params.kick_finetune_power = 2.; // test in steps of 5
  params.turn_finetune = 0;

  kick_params.tolerance_velocity = 0.3;//TOLERANCE_VELOCITY;
  kick_params.fast_tolerance_velocity = TOLERANCE_VELOCITY;
  kick_params.slow_tolerance_velocity = TOLERANCE_VELOCITY;
  kick_params.tolerance_direction = TOLERANCE_DIRECTION;
  kick_params.epsilon1 = EPSILON_1; // distance to kick margin
  kick_params.epsilon2 = EPSILON_2; // distance to player
  kick_params.angle_finesteps = 5.;
  kick_params.power_finesteps = 5.;
  kick_params.high_speed = 2.0;
  kick_params.slow_stop_threshold = 100.; // do not stop
  kick_params.medium_stop_threshold = 100.; // do not stop
  kick_params.fast_stop_threshold = 100.; // do not stop

  reset_statistics();  
}

/** destructor method */
Neuro_Kick::~Neuro_Kick(){
  for(int i=0;i<360*3;i+=RESOLUTION_DEG){ 
    delete nets[i];
  }
}
void Neuro_Kick::generate_actions(const Neuro_State& state){
  action_counter = 0;
  Value power, angle;

  /* add kick commands to action set */
  for(power=params.kick_power_min; power<=params.kick_power_max; power+=params.kick_power_inc){
    for(angle=params.kick_angle_min; angle<params.kick_angle_max; angle+=params.kick_angle_inc){
      actions[action_counter].unset_lock();
      actions[action_counter].unset_cmd();
      
      actions[action_counter].set_kick(power, angle);
      action_counter++;
    }
  }
}

void Neuro_Kick::generate_finetune_actions(const Neuro_State& state,const Cmd_Main& optimal_action){
  action_counter = 0;
  Value power, angle;
  Value opt_power, opt_angle;
  optimal_action.get_kick(opt_power, opt_angle);
  
  if(kick_params.angle_finesteps == 0.)
    kick_params.angle_finesteps = 1.;
  if(kick_params.power_finesteps == 0.)
    kick_params.power_finesteps = 1.;

  Value power_start = opt_power - params.kick_power_inc;
  Value power_end = opt_power + params.kick_power_inc;
  Value angle_start = opt_angle - params.kick_angle_inc;
  Value angle_end = opt_angle + params.kick_angle_inc;

  float fine_angle_inc = params.kick_angle_inc/kick_params.angle_finesteps;
  float fine_power_inc = params.kick_power_inc/kick_params.power_finesteps;

  for(power=power_start; power<=power_end; power+=fine_power_inc){
    for(angle=angle_start; angle<=angle_end; angle+=fine_angle_inc){
  
      actions[action_counter].unset_lock();
      actions[action_counter].unset_cmd();
      
      actions[action_counter].set_kick(power, angle);
      action_counter++;
    }
  }
}


bool Neuro_Kick::final_state(){
   
 
  if( !mdpInfo::is_ball_kickable_exact()){
 
    if(last_time_executed +1 == mdpInfo::mdp->time_current){
      get_neuro_state(current_state);
      if(is_success(current_state)){
	statistics.called++;
	statistics.succeeded++;
      }
      else {
	statistics.called++;
	statistics.failed++;
      }    
      //print_statistics();
      last_time_executed--;
    }

    return true;
  }
  return false;
}

void Neuro_Kick::reset_optimum(){
  minimum_direction_diff = kick_params.tolerance_direction;
  minimum_velocity_diff =  kick_params.tolerance_velocity;
  max_velocity = target_velocity - kick_params.tolerance_velocity;
}

/** if the costs are equal prefer actions wich lead to states with better velocity
    and direction accuracy.
*/ 
bool Neuro_Kick::is_optimum(const Neuro_State& state, float V, float Vmin){
  Value diff_direction, diff_velocity,velocity;
  Vector ball_vel = state.ball_vel;
  Value ang= 2*PI- state.my_angle;
  ball_vel.rotate(ang);

  diff_direction = Tools::get_abs_angle(ball_vel.arg() - target_direction);
  diff_velocity = fabs(ball_vel.norm() - target_velocity);
  velocity = state.ball_vel.norm();
  

  if(V<Vmin)
    return true; // V is actually better than best yet
  if(optimality_mode == VEL_PLUS){
    if(V==Vmin){ // maybe we can improve accuracy
      /*
      LOG_MOVE_D<<"Neuro Kick: is optimum velplus - testing vel "<<velocity<<" angle "
		<<RAD2DEG(state.ball_vel.arg())<<" diff dir "<<RAD2DEG(diff_direction)
		<<" current min_diff "<<minimum_direction_diff<<" vel diff "
		<<minimum_velocity_diff<<" max vel "<<max_velocity;
      */
      if((diff_direction < minimum_direction_diff) && // improve direction 
	 ((minimum_direction_diff>kick_params.tolerance_direction/2.) 
	  || (velocity >= max_velocity))){ // total improvement
	minimum_direction_diff = diff_direction;
	max_velocity = velocity; 
	return true;
      }
      if((diff_direction <= kick_params.tolerance_direction/2.) && // direction already good enough
	 (velocity > max_velocity)){  // better velocity. accuracy
	minimum_direction_diff = diff_direction;
	max_velocity = velocity; 
	return true;
      }
    }
    return false;
  }
  else{ // standard optimality mode
    if(V==Vmin){ // maybe we can improve accuracy
      /*
      LOG_MOVE_D<<"Neuro Kick: is optimum standard - testing vel "<<velocity<<" angle "
		<<RAD2DEG(state.ball_vel.arg())<<" diff dir "<<RAD2DEG(diff_direction)
		<<" current min_diff "<<RAD2DEG(minimum_direction_diff)<<" vel diff "
		<<minimum_velocity_diff<<" max vel "<<max_velocity;
      */
      if((diff_direction < minimum_direction_diff) && // improve direction 
	 ((minimum_direction_diff>kick_params.tolerance_direction/2.) 
	  || (diff_velocity <= minimum_velocity_diff))){ // total improvement
	minimum_direction_diff = diff_direction;
	minimum_velocity_diff = diff_velocity;
	return true;
      }
      if((diff_direction <= kick_params.tolerance_direction/2.) && // direction already good enough
	 (diff_velocity < minimum_velocity_diff)){  // better velocity. accuracy
	minimum_direction_diff = diff_direction;
	minimum_velocity_diff = diff_velocity;
	return true;
      }
    }
    return false;
  }
}

/** rule based pre-computing part of the neuro kick move  */ 
bool Neuro_Kick::decide_rule_based(Cmd & cmd){
 
  Vector rel_ball_pos;
  /*
  if(params.test_target_direction >= 0. && params.test_target_velocity >= 0.){ // do test
    target_direction = params.test_target_direction;
    target_velocity = params.test_target_velocity;
  }
  */
  
  if(kickmode == 6)
    target_direction = turn2_direction - mdpInfo::mdp->me->ang.v; // this is the desired rel. kick dir
 
  int x1,x2,x3,x4;
  x1=(int)mdpInfo::my_pos_abs().x;
  x2=(int)mdpInfo::my_pos_abs().y;
  Vector target;
  float target_dist = target_velocity * (1/(1-ServerOptions::ball_decay));
  target.init_polar(target_dist,(target_direction+mdpInfo::mdp->me->ang.v));
  x3 = x1 + (int)target.x;
  x4 = x2 + (int)target.y;

#ifdef LOG_KICK_ON
  LOG_DRAWLINE<<x1<<" "<<x2<<" "<<x3<<" "<<x4<<" red";
  LOG_MOVE_D<<"Neuro kick, Optimality Mode "<<optimality_mode<<" turn2dir: "<<RAD2DEG(turn2_direction);
#endif
  if (target_velocity > ServerOptions::ball_speed_max) 
    target_velocity = ServerOptions::ball_speed_max;
  
  sequence_terminated = false; // required only for compatibility with learning version
    
  //correct_ball_vel(current_state);

  if(last_time_executed +1 != mdpInfo::mdp->time_current){
    
    if(params.doku_level >= 1){
      printf("*********** New Kick Sequence started ***************\n");
      printf("kicking velocity: %g  directio: %g\n", target_velocity, target_direction);
    }
    if(params.doku_level >= 2){
      printf("Entering with state \n");
      print_state(current_state);
    }
    
    rel_ball_pos = current_state.ball_pos - current_state.my_pos;
    if(rel_ball_pos.norm() < ServerOptions::player_size){ //ball 'within' player: 
      //assume that ball is catched
      // try to put ball at 0.35, 0 in view direction
      
      if(params.doku_level >= 2){
	cout<<"NeuroKick: Assume that ball within player, kick before me\n";
      }      
      cmd.cmd_main.set_kick(35.0, 0.0);
      return false;
    }
  }
  else
    new_sequence_started = false;

  if((kickmode == 6) && (steps_done == 0)){ // turn and kick
    /*
    LOG_MOVE_D<<"Neuro Kick: abs turn "
	      <<Tools::get_abs_angle(mdpInfo::mdp->me->ang.v - turn2_direction);
    Value desired_turn = -(mdpInfo::mdp->me->ang.v - turn2_direction);
    LOG_MOVE_D<<"Neuro Kick: abs turn "
	      <<Tools::get_abs_angle(mdpInfo::mdp->me->ang.v - turn2_direction)
	      <<" direct "<<Tools::get_abs_angle(desired_turn);
    */
    if((Tools::get_abs_angle(mdpInfo::mdp->me->ang.v - turn2_direction) > 10/180. * PI) &&
       (mdpInfo::is_ball_kickable_next_time())){ // then turn first
      // do a corrected turn, code taken from move_basics.c
      double p_moment;
      //p_moment = desired_turn;
      p_moment = -(mdpInfo::mdp->me->ang.v - turn2_direction);
      p_moment = Tools::get_angle_between_mPI_pPI(p_moment);
      p_moment = p_moment * (1.0 + (ServerOptions::inertia_moment 
				    * (mdpInfo::my_vel_abs().norm())));
      if (p_moment > 3.14) p_moment = 3.14;
      if (p_moment < -3.14) p_moment = -3.14;
      p_moment = Tools::get_angle_between_null_2PI(p_moment);
      cmd.cmd_main.set_turn(p_moment);
#ifdef LOG_KICK_ON
      LOG_MOVE_D<<"Neuro kick, Trying to turn first by "<<RAD2DEG(p_moment);
#endif
      return false;
    }    
  }

  last_time_executed = mdpInfo::mdp->time_current;
  
  //ok everything is ok with the state we can use the neuro stuff!
  return true;
}
 
  
/** neuro kicking uses different nets for different target angles and target velocities
    this funcion selects the appropriate net with respect to the parameters target_angle and
    target_velocity
*/
Net* Neuro_Kick::choose_net(const Neuro_State&){
  int which_net;
  float resolution = RESOLUTION_DEG;
  float tmp;
  Net *selected_net;


  //printf("Choose Net: %g %g\n",target_velocity, target_direction);
  tmp = (target_direction/PI) *180 + (resolution/2.);
  if(tmp>=360) 
    tmp-=360;
  which_net = (int)(tmp/resolution);
  kick_params.tolerance_velocity = kick_params.slow_tolerance_velocity;
  if (target_velocity < 1.0){ // take slow speed nets if possible
    which_net += 2* (int)(360/RESOLUTION_DEG);
    if(which_net >= net_counter){ // the slow speed net is not available
      target_velocity = 1.0;
      
      if(params.doku_level >= 2)
	printf("NeuroKick: Sorry max speed net %d not available\n",which_net);
      
      which_net -= 2* (int)(360/RESOLUTION_DEG);
    }
  }  
  else if (target_velocity > kick_params.high_speed){ // take max speed nets if possible
    if(target_velocity >= 2.5)
      target_velocity = 2.5; // reasonable restriction
    kick_params.tolerance_velocity = kick_params.fast_tolerance_velocity;
    if((target_direction >2.61) && (target_direction <3.66)){
      // these directions require a  velocity tolerance of at least .4
      if(kick_params.tolerance_velocity < 0.4)
	kick_params.tolerance_velocity = 0.4;
    }
    which_net +=(int)(360/RESOLUTION_DEG);
    if(which_net >= net_counter){ // the max speed net is not available
      target_velocity =kick_params.high_speed;
      
      if(params.doku_level >= 2)
	printf("NeuroKick: Sorry max speed net %d not available\n",which_net);
      
      which_net -= (int)(360/RESOLUTION_DEG);
    }
  }  
  if((which_net>=0) && (which_net <net_counter)){
    //printf("Selected net: %d\n", which_net);
    selected_net = nets[which_net];
  }
  else {
    //if(selected_net->topo_data.in_count == 0){
    printf("NeuroKick: Sorry, no net known for this target\n");
    //return nets[0];
    return NULL;
  }

  /*
  if(params.doku_level >= DOKU_MEDIUM){
    printf("Neuro Shoot: rel dir: %.2f (+-%.2f) (%.2f)  vel: %.2f (+-%.2f) net: %d\n",
	   params.target_direction,params.tolerance_direction,
	   RAD2DEG(params.target_direction),
	   params.target_velocity,params.tolerance_velocity,which_net);
  }
  */
  return selected_net;
}

/** specifies the successful terminal states of the neuro kick move
    a state is a successful terminal state, if the ball has left the kick range
    in the right direction and with the right velocity. this conditions only have 
    be fullfilled within a certain tolerance region specified by tolerance_direction
    and tolerance_velocity
*/ 
bool Neuro_Kick::is_success(const Neuro_State& state){
  Value diff_direction, velocity;
  Vector ball_dist = state.ball_pos - state.my_pos;
  Vector ball_vel = state.ball_vel;

  Value ang= 2*PI- state.my_angle;
  ball_dist.rotate(ang);
  ball_vel.rotate(ang);
  
  if(ball_dist.norm() > ServerOptions::kickable_area -kick_params.epsilon1) {
    
    //compute difference between actual and target direction
    //diff_direction = fabs(ball_vel.arg() - target_direction);
    diff_direction = Tools::get_abs_angle(ball_vel.arg() - target_direction);

    //get velocity of the ball
    velocity = ball_vel.norm();
    
    //check velocity and direction
    if(optimality_mode == VEL_PLUS){ // try to get target_velocity or more
      if((diff_direction < kick_params.tolerance_direction) &&
	 (velocity > target_velocity - kick_params.tolerance_velocity)){ // 
	// successful!
	return true;
      }
      return false;
    }
    else{ // STANDARD
      if((diff_direction < kick_params.tolerance_direction) &&
	 (fabs(velocity - target_velocity) < kick_params.tolerance_velocity)){ // velocity o.k.
	// successful!
	return true;
      }
      return false;
    }
  }
  return false;
}

/** specifies the unsuccessful terminal states of the neuro kick move
    a state is an unsuccsesful terminal state, if the ball has left the kick range
    but is no successful terminal state.
*/ 
bool Neuro_Kick::is_failure(const Neuro_State& state){
  Vector ball_dist = state.ball_pos - state.my_pos;

  if(ball_dist.norm() > ServerOptions::kickable_area - kick_params.epsilon1) {
    //left the kickrange - so its the negation of is_success()
    return !(is_success(state));
  }
  
  if(ball_dist.norm() < ServerOptions::player_size+0.15){
    return true;
  }


  return false;
}


/** neuro kick just uses relative ball position and ball velocity as
    network input.
    attention: make sure to have the used net at least 4 input neurons!
*/ 
void Neuro_Kick::get_features(const Neuro_State& state, float *net_in){
  
  //get interesting info from state
  Vector rel_ball_pos = state.ball_pos - state.my_pos;
  Vector ball_vel = state.ball_vel;
  
  //turn coordinate system such that current viewangle = 0
  Value ang= 2*PI- state.my_angle;
  rel_ball_pos.rotate(ang);
  ball_vel.rotate(ang);
  
  //feed net
  net_in[0] = rel_ball_pos.x;
  net_in[1] = rel_ball_pos.y;
  net_in[2] = ball_vel.x;
  net_in[3] = ball_vel.y;
}
void Neuro_Kick::print_state(const Neuro_State& state){
  //get interesting info from state
  Vector rel_ball_pos = state.ball_pos - state.my_pos;
  Vector ball_vel = state.ball_vel;
  
  //turn coordinate system such that current viewangle = 0
  Value ang= 2*PI- state.my_angle;
  //rel_ball_pos.rotate(ang);
  ball_vel.rotate(ang);

  printf("AGENT (%.2f, %.2f) ball_vel (%.2f, %.2f) \n", 
	 rel_ball_pos.x, rel_ball_pos.y, ball_vel.x, ball_vel.y );

} 
