#include "neuro_kick2.h"
#include "mdp_info.h"
#include "options.h"
#include "move_factory.h"
#include "tools.h"
#include "logger.h"

//#define LOG_KICK2_ON

Neuro_Kick2::Neuro_Kick2(float target_vel = 2.0, float target_dir = 0.0)
{ 
  /* copy arguments of neuro kick 2 move */
  target_velocity = target_vel;
  target_direction = target_dir;

  use_new_kicks = true;

  /* init some other stuff */
  last_time_executed = -1;

  /* load net for neuro kick 2 move */
  action_counter = 0;
  net_counter = 0;

  if(use_new_kicks == true){
    nets[0] = new Net();
    if(nets[0]->load_net("./data/nets_neuro_kick2/ultrakicks.net") == FILE_ERROR){
      cout<<"NeuroKick2: No net file found...\n";
      exit(0);
    }
    
    nets[1] = new Net();
    if(nets[1]->load_net("./data/nets_neuro_kick2/hardkicks.net") == FILE_ERROR){
      cout<<"NeuroKick2: No net file found...\n";
      exit(0);
    }
  }
  else{
    nets[0] = new Net();
    if(nets[0]->load_net("./data/nets_neuro_kick2/kick2_2.5.net") == FILE_ERROR){
      cout<<"NeuroKick2: No net file found...\n";
      exit(0);
    }
    
    nets[1] = new Net();
    if(nets[1]->load_net("./data/nets_neuro_kick2/kick2_2.0.net") == FILE_ERROR){
      cout<<"NeuroKick2: No net file found...\n";
      exit(0);
    }
  }
  
  net_counter = 2;

  /* set neuro kick params */
  params.costs_failure = 1.0;
  params.costs_success = 0.0;
  params.doku_level = 0;  // 0 = silence
  
  /* neuro kick uses no dashs */
  params.dash_power_min = 0;
  params.dash_power_inc = 0;
  params.dash_power_max = 0;
  
  /* neuro kick uses no turns */
  params.turn_angle_min = 0;
  params.turn_angle_inc = 0;
  params.turn_angle_max = 0;

  /* neuro kick uses kicks */
  params.kick_power_min = 10;
  params.kick_power_inc = 10;
  params.kick_power_max = 100;
  params.kick_angle_min = 0;
  params.kick_angle_inc = 2*PI/45.;
  params.kick_angle_max = 2*PI;


  params.kick_finetune = 2*PI/360.;
  params.kick_finetune_power = 1;
  params.turn_finetune = 0;

 
  /* set custom neuro_kick2 params */
  tolerance_velocity = 0.2;
  tolerance_direction = 0.05;
  min_ball_dist = 0.15;
  kickable_tolerance = 0.15;
  
  reset_statistics();
}

Neuro_Kick2::~Neuro_Kick2(){
  for(int i=0;i<net_counter;i++){
    delete nets[i];
  }
}

/** rule based pre-computing part of the neuro kick move  */ 
bool Neuro_Kick2::decide_rule_based(Cmd & cmd){
 
  
  int x1,x2,x3,x4;
  x1=(int)mdpInfo::my_pos_abs().x;
  x2=(int)mdpInfo::my_pos_abs().y;
  Vector target;
  float target_dist = target_velocity * (1/(1-ServerOptions::ball_decay));
  target.init_polar(target_dist,(target_direction));
  x3 = x1 + (int)target.x;
  x4 = x2 + (int)target.y;

  LOG_DRAWLINE<<x1<<" "<<x2<<" "<<x3<<" "<<x4<<" magenta";
    
  if(target_velocity > 2.3)
    tolerance_velocity = 0.4;
  else
    tolerance_velocity = 0.2;

  sequence_terminated = false; // required only for compatibility with learning version
  
  if(last_time_executed +1 != mdpInfo::mdp->time_current){
#ifdef LOG_KICK2_ON
    LOG_MOVE << "Neuro_Kick2: Starting kick - vel: " << target_velocity << " dir: " << target_direction << ".";
#endif
  }
  
  /*
  Vector ball_dist = current_state.ball_pos-current_state.my_pos;
  if(ball_dist.norm() <= 0.4){
    LOG_MOVE << "ball dist <= 0.4 -> collision?!? set ball vel to (0,0) ";
    current_state.ball_vel = Vector(0,0);
  }
  */

  new_sequence_started = false;
  last_time_executed = mdpInfo::mdp->time_current;
  
  //ok everything is ok with the state we can use the neuro stuff!
  return true;
}
/** specifies the successful terminal states of the neuro kick move
    a state is a successful terminal state, if the ball has left the kick range
    in the right direction and with the right velocity. this conditions only have 
    be fullfilled within a certain tolerance region specified by tolerance_direction
    and tolerance_velocity
*/ 
bool Neuro_Kick2::is_success(const Neuro_State& state){
  float ball_dist = (state.ball_pos - state.my_pos).norm();
  
  if(ball_dist > ServerOptions::kickable_area - kickable_tolerance) { 
    // ball has left my Kickrange

    // compute difference between real and target direction
    float diff_direction = state.ball_vel.arg() - target_direction; 
    diff_direction = fabs(Tools::get_angle_between_mPI_pPI(diff_direction));
    
    // compute velocity of the ball
    float diff_velocity = fabs(state.ball_vel.norm() - target_velocity); 

    //check velocity and direction
    if((diff_direction < tolerance_direction) 
       && (diff_velocity < tolerance_velocity)){ 
      // successful!
      return true;
    }
  }
  
  //no successful state
  return false;
}

/** specifies the unsuccessful terminal states of the neuro kick move
    a state is an unsuccsesful terminal state, if the ball has left the kick range
    but is no successful terminal state.
*/ 
bool Neuro_Kick2::is_failure(const Neuro_State& state){
  Vector ball_dist = state.ball_pos - state.my_pos;
  /*
  float pos_angle = ball_dist.arg();
  float vel_angle = state.ball_vel.arg();
  
  if((Tools::get_abs_angle(pos_angle-vel_angle) < PI/2)&&(state.ball_vel.norm() > 0.1)){
    if(ball_dist.norm() > ServerOptions::kickable_area - (kickable_tolerance+0.1)) {
      //left the kickrange - so its the negation of is_success()
      return !(is_success(state));
    }
  }
  */

  if(ball_dist.norm() > ServerOptions::kickable_area - kickable_tolerance) {
    //left the kickrange - so its the negation of is_success()
    return !(is_success(state));
  }
  
  if(ball_dist.norm() < ServerOptions::player_size+min_ball_dist){
    return true;
  }

  /*
  for(int i=1;i<=11;i++){
    if(mdpInfo::is_opponent_pos_valid(i)){
      Vector op_pos = mdpInfo::opponent_pos_abs(i);
      if(op_pos.distance(state.ball_pos) < ServerOptions::kickable_area){
	return true;
      }
    }
  }
  */

  return false;
}
 
void Neuro_Kick2::generate_actions(const Neuro_State& state){
  action_counter = 0;
 
  float turn_angle = target_direction - state.my_angle;  
  turn_angle = Tools::get_angle_between_mPI_pPI(turn_angle);
  
  turn_angle = turn_angle*(1.0 + state.my_vel.norm()*ServerOptions::inertia_moment); 
  if(fabs(turn_angle > PI)){
    if(turn_angle < 0){
      turn_angle = -PI;
    } 
    else {
      turn_angle = PI;
    } 
  }
  
  actions[action_counter].unset_lock();
  actions[action_counter].unset_cmd();
  actions[action_counter].set_turn(turn_angle);
  action_counter++;

  generate_kicks(state,params.kick_power_min , params.kick_power_max, params.kick_power_inc, 0, 2*PI,params.kick_angle_inc);
 
}

/** implementation of the Abstract_Move interface final_state() 
    returning true indicates the strategy component to select a new move.
*/ 
bool Neuro_Kick2::final_state(){
  if( !mdpInfo::is_ball_kickable_exact()){
#ifdef LOG_KICK2_ON
    LOG_MOVE << "Neuro_Kick2: Ball has left the kick range. terminate. ";
#endif
    last_time_executed = -1;
    return true;
  }

  return false;
}


/** neuro kicking uses different nets for different target angles and target velocities
    this funcion selects the appropriate net with respect to the parameters target_angle and
    target_velocity
*/
Net* Neuro_Kick2::choose_net(const Neuro_State&){
  if(target_velocity > 2.1){
    return nets[0];
  }
  if(target_velocity > 1.6){
    return nets[1];
  }

  return nets[1];
}


/** neuro moves are using a greedy policy wrt. to the neural represented value function.
    if the costs of two possible  successor states are equal the greedy policy could choose
    one successor state by chance. Overloading is_optimum(...) allows to select the successor state
    by aditional criteria. 
    for neuro kick we prefer states with better direction and velocity accuracy. 
*/ 
/** if the costs are equal prefer actions wich lead to states with better velocity
    and direction accuracy.
*/ 
bool Neuro_Kick2::is_optimum(const Neuro_State& state, float V, float Vmin){
  Value diff_direction, diff_velocity,velocity;
  Vector ball_vel = state.ball_vel;
  //Value ang= 2*PI- state.my_angle;
  //ball_vel.rotate(ang);

  diff_direction = Tools::get_abs_angle(ball_vel.arg() - target_direction);
  diff_velocity = fabs(ball_vel.norm() - target_velocity);
  velocity = state.ball_vel.norm();
  

  if(V<Vmin)
    return true; // V is actually better than best yet
  if(optimality_mode == VEL_PLUS){
    if(V==Vmin){ // maybe we can improve accuracy
      /*
      LOG_MOVE_D<<"Neuro Kick: is optimum velplus - testing vel "<<velocity<<" angle "
		<<RAD2DEG(state.ball_vel.arg())<<" diff dir "<<RAD2DEG(diff_direction)
		<<" current min_diff "<<minimum_direction_diff<<" vel diff "
		<<minimum_velocity_diff<<" max vel "<<max_velocity;
      */
      if((diff_direction < minimum_direction_diff) && // improve direction 
	 ((minimum_direction_diff>tolerance_direction/2.) 
	  || (velocity >= max_velocity))){ // total improvement
	minimum_direction_diff = diff_direction;
	max_velocity = velocity; 
	return true;
      }
      if((diff_direction <= tolerance_direction/2.) && // direction already good enough
	 (velocity > max_velocity)){  // better velocity. accuracy
	minimum_direction_diff = diff_direction;
	max_velocity = velocity; 
	return true;
      }
    }
    return false;
  }
  else{ // standard optimality mode
    if(V==Vmin){ // maybe we can improve accuracy
      /*
      LOG_MOVE_D<<"Neuro Kick: is optimum standard - testing vel "<<velocity<<" angle "
		<<RAD2DEG(state.ball_vel.arg())<<" diff dir "<<RAD2DEG(diff_direction)
		<<" current min_diff "<<RAD2DEG(minimum_direction_diff)<<" vel diff "
		<<minimum_velocity_diff<<" max vel "<<max_velocity;
      */
      if((diff_direction < minimum_direction_diff) && // improve direction 
	 ((minimum_direction_diff>tolerance_direction/2.) 
	  || (diff_velocity <= minimum_velocity_diff))){ // total improvement
	minimum_direction_diff = diff_direction;
	minimum_velocity_diff = diff_velocity;
	return true;
      }
      if((diff_direction <= tolerance_direction/2.) && // direction already good enough
	 (diff_velocity < minimum_velocity_diff)){  // better velocity. accuracy
	minimum_direction_diff = diff_direction;
	minimum_velocity_diff = diff_velocity;
	return true;
      }
    }
    return false;
  }
}

float Neuro_Kick2::default_V(const Neuro_State& state){
  
  float ball_dist = (state.ball_pos - state.my_pos).norm();

  if(ball_dist >  ServerOptions::kickable_area) return 1.0;
  if(ball_dist <  ServerOptions::player_size+min_ball_dist) return 1.0;
  
  return ball_dist/2.0;
}

void Neuro_Kick2::reset_optimum(){
 
  minimum_direction_diff = tolerance_direction;
  minimum_velocity_diff =  tolerance_velocity;  
  max_velocity = target_velocity - tolerance_velocity;

}

/** neuro kick just uses relative ball position and ball velocity as
    network input.
    attention: make sure to have the used net at least 4 input neurons!
*/ 
void Neuro_Kick2::get_features(const Neuro_State& state, float *net_in){
  if(use_new_kicks == true){

    /* relative features version */
    Vector pos = state.ball_pos-state.my_pos;
    Vector vel = state.ball_vel;
    Value view = state.my_angle;
    
    /* rotate whole system (pos.y should be zero*/
    pos.rotate(2*PI - target_direction);
    vel.rotate(2*PI - target_direction);
    view = Tools::get_angle_between_mPI_pPI((view - target_direction)-PI);
    
    if(view<0) {
      vel.y = -vel.y;
      pos.y = -pos.y;
    }
    
    net_in[0] = pos.x; //distance to target
    net_in[1] = pos.y; //velocity in x direction
    net_in[2] = vel.x; //distance to target
    net_in[3] = vel.y; //velocity in x direction
    net_in[4] = fabs(view); //abs. relative view angle to target
  }
  else{
    Vector my_ball_dist = state.ball_pos - state.my_pos;
    float my_ball_angle = my_ball_dist.arg() - state.my_angle;
    my_ball_angle = Tools::get_angle_between_mPI_pPI(my_ball_angle);
    
    float my_view_angle =  target_direction - state.my_angle;
    my_view_angle = Tools::get_angle_between_mPI_pPI(my_view_angle);
    
    Value ang= 2*PI- target_direction;
    Vector bvel = state.ball_vel;
    Vector mvel = state.my_vel;
    bvel.rotate( ang  );
    mvel.rotate( ang );
    my_ball_dist.rotate( ang );
    
    net_in[0] = my_ball_dist.x;
    net_in[1] = my_ball_dist.y;
    net_in[2] = bvel.x;
    net_in[3] = bvel.y;
    net_in[4] = my_view_angle;
  }
}

void Neuro_Kick2::print_state(const Neuro_State& state){
  //get interesting info from state
  Vector rel_ball_pos = state.ball_pos - state.my_pos;
  Vector ball_vel = state.ball_vel;
  
  //turn coordinate system such that current viewangle = 0
  Value ang= 2*PI- state.my_angle;
  //rel_ball_pos.rotate(ang);
  ball_vel.rotate(ang);

printf("AGENT (%.2f, %.2f) ball_vel (%.2f, %.2f) \n", 
	 rel_ball_pos.x, rel_ball_pos.y, ball_vel.x, ball_vel.y );

} 
