#ifndef NEURO_MOVE_H_
#define NEURO_MOVE_H_

#include "Vector.h"
#include "globaldef.h"
#include "splayer.h"
#include "n++.h"
#include "base_moves.h"

#define ANGLE_MAX 2*PI
#define POWER_MAX 100


#define MAX_NETS 100
#define MAX_ACTIONS 1000

/** this structure store a local section of the global mdp state which is of interest to
    the neuro move
*/ 
struct Neuro_State{
  Vector my_vel;
  Vector my_pos;
  float my_angle;
  
  Vector ball_vel;
  Vector ball_pos;
  
  Vector opnt1_pos;
  Vector opnt1_vel;
};


/** base class for all neuro moves */ 
class Neuro_Move : public Main_Move {

 protected:
  bool sequence_terminated;

  /** actual_net is a reference to the net currently used to calculate the value function */ 
  Net *actual_net;

  /** list of references to all nets used by the neuro move */ 
  Net *nets[MAX_NETS];
  
  /** number of nets actually used by the neuro move*/ 
  int net_counter;

  /** all possible actions that can be performed in the current state*/ 
  Cmd_Main actions[MAX_ACTIONS];

  /** number of actions actually used in a current state*/ 
  int action_counter;

  /** neuro params*/ 
  struct {
    /** documentation level */
    int doku_level;
    /** costs that will occure if the system is running into a successful final state */ 
    Value costs_success;
    /** costs that will occure if the system is running into an unsuccessful final state */ 
    Value costs_failure;
    
    /** action resolution params */
    Value dash_power_min;
    Value dash_power_inc;
    Value dash_power_max;

    Value turn_angle_min;
    Value turn_angle_inc;
    Value turn_angle_max;
  
    Value turn_finetune;


    Value kick_power_min;
    Value kick_power_inc;
    Value kick_power_max;
    
    Value kick_angle_min;
    Value kick_angle_inc;
    Value kick_angle_max;

    Value kick_finetune;
    Value kick_finetune_power;

  
  } params;

 
  /** current state*/ 
  Neuro_State current_state;
  
  /** used to decide if a move started
      set last_time_executed to -1 in the constructor of your move
  */
  int last_time_executed;
  

  /* prediction related stuff */
  Cmd_Main prediction_cmd;
  Neuro_State prediction_state;
  int prediction_time;

  /** execute the neuro move.
   */
  virtual void execute(Cmd &);
  
  /** copies a small part of the state from the mdp to an Neuro_State variable */
  void get_neuro_state(Neuro_State& state);
 
  /** this method is called by execute() before the decide_neural() method and allows to enrich the
      move by some rulebased considerations. For example to check if the current state can be 
      handeled by the neural component at all. if not, one could set a command by hand here and return
      false to indicate this to execute(). On the other hand returning true indicates that 
      decide_rule_based() doesnt want to execute a command and the neuro stuff can be started
  */ 
  virtual bool decide_rule_based(Cmd & cmd){return true;}
  
  virtual void check_decision(Cmd & cmd){;}

  /** this method selects the action leading to the optimal successor state with respect to 
      the value function represented by the neural network. */ 
  void decide_neural(Cmd & cmd);
 
  /** this method selects the action leading to the optimal successor state with respect to 
      the value function represented by the neural network. */ 
  Cmd_Main decide_neural_finetune(Cmd_Main & cmd_main, float Vmin_old);
  
  /** checks if a given state is a successful terminal state*/ 
  virtual bool is_success(const Neuro_State& state)=0;
  
  /** checks if a given state is a unsuccessful terminal state*/ 
  virtual bool is_failure(const Neuro_State& state)=0;
  
  /** choose an adequate net with respect to the current state. This method will be called
      everytime the move is executed.
      Overload this method if your neuro move needs more that one net.
  */
  virtual Net* choose_net(const Neuro_State& state){return nets[0];}
  
  /** this method defines the optimal state criterion. */ 
  virtual bool is_optimum(const Neuro_State& predicted_state, float V, float Vmin);
   
  /** resets the current optimal values - belongs to is_optimum */ 
  virtual void reset_optimum(){};  // default do nothing
   
  /** this method generates a set of allowed actions for a given state. the actions
      have to be stored in the actions[] array and the number of generated actions
      has to written to action_counter.
  */ 
  virtual void generate_actions(const Neuro_State&){action_counter = 0;}


  void generate_turns(const Neuro_State&, Value root_angle, Value max_angle, Value inc_angle);

  void generate_dashes(const Neuro_State&, Value min_power, Value max_power, Value inc_power);

  void generate_kicks(const Neuro_State&, Value min_power, Value max_power, Value inc_power, 
		      Value root_angle, Value max_angle, Value inc_angle);
  
  /** this method generates a set of allowed fine tune actions for a given state and a given
      optimal action. the new fine tune actions again  have to be
      stored in the actions[] array and the number of generated fine tune actions actions
      has to written to action_counter. if you don`t want to use fine tuning just don`t override
      generate_fine_tune_actions;
  */ 
  virtual void generate_finetune_actions(const Neuro_State&, const Cmd_Main& opt_action){action_counter = 0;}
  
  /** extracts parts of an internal state which serve as features for the neuro move 
      Overload this method to specify the features for your neuro move*/ 
  virtual void get_features(const Neuro_State& state, float *net_input)=0;
  
  /** these method calculates the value function for a given state.by testing if the state is
      terminal or otherwise calling comute_VNN */
  float compute_V(const Neuro_State &state);

  /** these method calculates the value function for a given state with the actual_net neural network*/
  float compute_VNN(const Neuro_State &state);
  
  /** these methods provide a one step model*/
  void model(const Neuro_State& old_state, const Cmd_Main &, Neuro_State& predicted_state, float = 0, float = 0);
  
  /** if all actions lead to failure wrt. the compute_V value function the best action wrt. to default_V ist selected */
  virtual float default_V(const Neuro_State&){return 1.0;} 

  virtual void print_state(const Neuro_State&){}; 

  /* returns true if a move has started in the current time step 
     this implementation is wrong if a move ends in time step t and 
     is called immediately again in time step t+1 
  */
  //bool move_started(){return !(last_time_executed+1 == mdpInfo::mdp->time_current);}

 public:

  virtual inline ~Neuro_Move(){}
  void print_statistics();
  void get_state(char *buffer, const Neuro_State& state);
  void get_action(char *buffer, const Cmd_Main &cmd);
};

#endif
