#include "neuro_stop_ball.h"
#include "mdp_info.h"
#include "options.h"
#include "tools.h"
#include "move_factory.h"
#include "logger.h"

/** constructor method */
Neuro_Stop_Ball::Neuro_Stop_Ball(float p_max_velocity=0.7){
  max_velocity = p_max_velocity;
  max_ball_dist = 0.7;

  char *netname = "./data/nets_neuro_stop/stop_ball_05.net";
  
  // important: do init first, since net loading might be aborted
  nets[0] = new Net();
  if(nets[0]->load_net(netname) == FILE_ERROR){
    cout<<"NeuroStopBall: No net-file found in directory ./nets - stop loading\n";
    return;
  }
  
  /* set neuro_stop_ball params */
  params.costs_failure = 1.0;
  params.costs_success = 0.0;
  params.doku_level = 0;  // 0 = silence
  
  /* neuro_stop_ball uses no dashs */
  params.dash_power_min = 0;
  params.dash_power_inc = 0;
  params.dash_power_max = 0;
  
  /* neuro_stop_ball uses no turns */
  params.turn_angle_min = 0;
  params.turn_angle_inc = 0;
  params.turn_angle_max = 0;

  /* neuro_stop_ball kicks */
  params.kick_power_min = 10;
  params.kick_power_inc = 10;
  params.kick_power_max = 100;
  params.kick_angle_min = 0;
  params.kick_angle_inc = 2*PI/20;
  params.kick_angle_max = 2*PI;

  

}

/** destructor method */
Neuro_Stop_Ball::~Neuro_Stop_Ball(){
  delete nets[0];
}

void Neuro_Stop_Ball::generate_actions(const Neuro_State& state){
  action_counter = 0;
  double power, angle;
  
  /* add kick commands to action set */
  for(power=params.kick_power_min; power<=params.kick_power_max; power+=params.kick_power_inc){
    for(angle=params.kick_angle_min; angle<params.kick_angle_max; angle+=params.kick_angle_inc){
      actions[action_counter].unset_lock();
      actions[action_counter].unset_cmd();
      
      actions[action_counter].set_kick(power, angle);
      action_counter++;
    }
  }
}

/** specifies the successful terminal states of the neuro stop ball move

    stop ball is finished successful if the ball is inside the 
    players kick range and both, player and ball are slow enough
*/ 
bool Neuro_Stop_Ball::is_success(const Neuro_State& state){
  
  if(mdpInfo::my_distance_to_ball() <= max_ball_dist){
    if(state.ball_vel.norm()< max_velocity){
      return true;
    }
  }
  
  return false;
}

/** specifies the unsuccessful terminal states of the neuro stop ball move
    
 */ 
bool Neuro_Stop_Ball::is_failure(const Neuro_State& state){
  if(mdpInfo::is_ball_kickable_exact() == false){
    return true;
  }
  return false;
}
/** features of neuro stop ball:
    
    (0) relative ball position x (rotated)
    (1) relative ball position y (rotated)
    (2) player velocity x-direction (rotated)
    (3) player velocity y-direction (rotated)
    (4) ball velocity x-direction (rotated)
    (5) ball velocity y-direction (rotated)
*/ 
void Neuro_Stop_Ball::get_features(const Neuro_State& state, float *net_in){

  Vector my_ball_dist = state.ball_pos - state.my_pos;
  Vector ball_vel = state.ball_vel;
  Vector my_vel = state.my_vel;
   
  float ball_angle = my_ball_dist.arg() - state.my_angle;
  ball_angle = Tools::get_angle_between_mPI_pPI(ball_angle);
  
  Value ang = 2*PI- state.my_angle;
  ball_vel.rotate( ang );
  my_ball_dist.rotate( ang );
  my_vel.rotate( ang );
  
  net_in[0] = my_ball_dist.x;
  net_in[1] = my_ball_dist.y;
  net_in[2] = my_vel.x;
  net_in[3] = my_vel.y;
  net_in[4] = ball_vel.x;
  net_in[5] = ball_vel.y;

  net_in[6] = ball_angle;
  if(my_ball_dist.norm() < ServerOptions::kickable_area){
    net_in[7] = 1.0;
  }
  else {
    net_in[7] = 0.0;
  }

}


void Neuro_Stop_Ball::print_state(const Neuro_State& state){

  float features[6];
  get_features(state, features);
  
  printf("StopBall step: %d:  ball_dist %.2f  ball_angle %.2f  ball_vel (%.2f %.2f) \n",
	 mdpInfo::mdp->time_current, features[0],  features[1],  features[4],  features[5]);
}

bool Neuro_Stop_Ball::final_state(){
 
  if(mdpInfo::is_ball_kickable_exact() == true){
    if(mdpInfo::ball_vel_abs().norm()< max_velocity){
      LOG_MOVE << "Neuro_Stop_Ball: ball slowed down!";
      return true;
    }
  }
  if(mdpInfo::is_ball_kickable_exact() == false){
    LOG_MOVE << "Neuro_Stop_Ball: ball has left the kickrange!";
    return true;
  }
  
  return false;
}
  
