/** \file base_policies.h
    The base policy interfaces are defined here
    What is a policy ?
    A policy is a decision making instance which returns a most approriate move
    given a certain world model state. The strategy decides which policy is
    to be used according to the playmode (or other parameters)
    It is sensible to use config files for each policy. Since policies are 
    sometimes expensive objects to create, they are managed in an object
    factory.
*/
#ifndef _BASE_POLICY_H_
#define _BASE_POLICY_H_

#include "base_moves.h"
#include "policy_defs.h" 
#include "move_factory.h"

class Abstract_Policy{
 public:
  /** get the unique id of the policy.
      \return policy id.
  */ 
  inline virtual int id() {return NO_POLICY;}
  /** The main decision making instance.
   \return the move that is to be executed
  */
  virtual Abstract_Move *playon()=0;
};

/** Main policy base class.
    You have to overload playon().
*/
class Main_Policy : public Abstract_Policy{
public:
  /** get the unique id of the policy.
      \return policy id.
  */ 
  inline virtual int id() {return NO_POLICY;}
  /** The main decision making instance.
   \return the move that is to be executed
  */
  virtual Main_Move *playon(){return Move_Factory::get_Main_Move();}
  virtual Main_Move *reconsider_move(Main_Move *current_move)
    {return current_move;} //default: do not interrupt current move
};


/** Neck policy base class.
    You have to overload playon().
*/
class Neck_Policy : public Abstract_Policy{
public:
  /** get the unique id of the policy.
      \return policy id.
  */ 
  inline virtual int id() {return NO_POLICY;}
  /** The main decision making instance.
   \return the move that is to be executed
  */
  virtual Neck_Move *playon(){return Move_Factory::get_Neck_Move();}
};


/** Say policy base class.
    You have to overload playon().
*/
class Say_Policy : public Abstract_Policy{
public:
  /** get the unique id of the policy.
      \return policy id.
  */ 
  inline virtual int id() {return NO_POLICY;}
  /** The main decision making instance.
   \return the move that is to be executed
  */
  virtual Say_Move *playon(){return Move_Factory::get_Say_Move();}
};


/** Say policy base class.
    You have to overload playon().
*/
class View_Policy : public Abstract_Policy{
public:
  /** get the unique id of the policy.
      \return policy id.
  */ 
  inline virtual int id() {return NO_POLICY;}
  /** The main decision making instance.
   \return the move that is to be executed
  */
  virtual View_Move *playon(){return Move_Factory::get_View_Move();}
};

#endif //_BASE_POLICY_H_
