/** \file base_strategies.h
    The base strategy classes are defined here.
    What is a strategy ?
    A strategy in this case is an instance whose job it is to select an appropriate 
    policy for standard and global situations. So you might have a goalie strategy
    which during normal play uses a neural policy and for standard situations like
    corner kick uses some rulebased policy for this standard situation. The strategy 
    can also be dynamically exchanged during gameplay. So if you are leading by 5-0, 
    maybe you want all your players, including the goalie, to use some offensive 
    strategy.
*/
#ifndef _BASE_STRATEGY_H_
#define _BASE_STRATEGY_H_

#include "base_policies.h"
#include "base_moves.h"
#include "strategy_defs.h"
#include "policy_factory.h"

/** The base class for all strategies. */
class Abstract_Strategy{
protected:
  /** Overload this method to implement your own policy selection strategy. 
      \return a policy of your choice 
 */
  virtual Abstract_Policy *select_policy()=0;
public:
  /** This method sets the policy to what is returned by select_policy.*/
  virtual void update_policy()=0;
  /** get the current policy. */
  inline virtual Abstract_Policy *get_policy()=0;
  /** get the move returned by policy->playon() */
  inline virtual Abstract_Move *get_move()=0;
  inline virtual int id(){return NO_STRATEGY;}
};

/** the base main strategy. */
class Main_Strategy : public Abstract_Strategy {
protected:
  Main_Policy *main_policy;
  virtual Main_Policy *select_policy(){return Policy_Factory::get_Main_Policy();}
public:
  /** This method sets the policy to what is returned by select_policy.*/
  inline virtual void update_policy(){main_policy = select_policy();}
  inline virtual Main_Policy *get_policy(){return main_policy;}
  inline virtual Main_Move *get_move(){return main_policy->playon();}
  inline virtual Main_Move *reconsider_move(Main_Move* current_move) {return main_policy->reconsider_move(current_move);}
};

class Neck_Strategy : public Abstract_Strategy {
protected:
  Neck_Policy *neck_policy;
  virtual Neck_Policy *select_policy(){return Policy_Factory::get_Neck_Policy();}
public:
  /** This method sets the policy to what is returned by select_policy.*/
  inline virtual void update_policy(){neck_policy = select_policy();}
  inline virtual Neck_Policy *get_policy(){return neck_policy;}
  inline virtual Neck_Move *get_move(){return neck_policy->playon();}
};

class Say_Strategy : public Abstract_Strategy {
protected:
  Say_Policy  *say_policy;
  virtual Say_Policy *select_policy(){return Policy_Factory::get_Say_Policy();}
public:
  /** This method sets the policy to what is returned by select_policy.*/
  inline virtual void update_policy(){say_policy = select_policy();}
  inline virtual Say_Policy *get_policy(){return say_policy;}
  inline virtual Say_Move *get_move(){return say_policy->playon();}
};

class View_Strategy : public Abstract_Strategy {
protected:
  View_Policy *view_policy;
  virtual View_Policy *select_policy(){return Policy_Factory::get_View_Policy();}
public:
  /** This method sets the policy to what is returned by select_policy.*/
  inline virtual void update_policy(){view_policy = select_policy();}
  inline virtual View_Policy *get_policy(){return view_policy;}
  inline virtual View_Move *get_move(){return view_policy->playon();}
};


#endif //_BASE_STRATEGY_H_
