#ifndef _POLICY_TOOLS_H
#define _POLICY_TOOLS_H

#include "globaldef.h"
#include "Vector.h"


#define DEFENSE_AREA_X 16.0
#define DEFENSE_AREA_Y 16.0
#define DEFENSE_OFFSET 8.0

/*****************************************************************************/

struct Pass{
  int player_num;
  Vector target;
  float velocity;
};

/*****************************************************************************/

struct Go2Ball_Steps {
  const static int MY_TEAM = 1;
  const static int THEIR_TEAM = -1;
  const static int NO_TEAM = 0;
  
  int number;
  int side;
  float steps;
};
/*****************************************************************************/
typedef Vector Point;
  
class Intercept {
  const static int MAX_STEPS= 14; //2**(14-2)= 4096 > 3000
  bool ball_in_player_radius_at_time( Value ball_vel_decay_to_the_power_of_time, int time );
 public:
  /** ball_pos[t]= ball_pos + 1.0*ball_vel + ball_vel_decay*ball_vel + ... + ball_vel_decay**(t-1) * ball_vel
      || player_pos[t] - player_pos || =  t * player_vel_max
      
      || p || denotes the euclidean norm of the vector p  ;-)
      We seek the first time t, so that ball_pos[t] is in reach of the player.
      
  */
  Point  ball_pos;
  Vector ball_vel;
  Value  ball_vel_decay;
  Point  player_pos;
  Value  player_vel_max;
  /** player_radius is the radius of the player, in which he can catch the ball at time 0.
      Negative Values are allowed, and can be interpreted as additional safety distance 
      (this feature will be probably never used, but anyway :-)*/
  Value  player_radius;
 public:
  /** if the needed time to intercept ball is > 2**(MAX_STEPS-2)  then 
      2**(MAX_STEPS-2) is the returned time and pos ist the ball position  
      after 2**(MAX_STEPS-2) steps */
  void minimal_time_player_reaches_ball_at_pos( int & time, Point & pos );
};


/*****************************************************************************/

/** The class MachtPositions implements am matching mechanism. 
    It can match up ton num_players_MAX to the maximum of num_targets_MAX 
    of targets. Both players and targets are represented as vectors.

    The matching mechnism does use a heuristic. It doesn't consider all
    possible matchings, but instead every target can just be matched by
    one of the nearest 3 players.

    if num_targets_MAX=6, then ther are
    3^num_targets_MAX= 3^6= 729  possible solutions
    
    You can also use more then 3 nearest neighbours, but the code is optimized
    just for 3, so you will have to change some details (i.e. the computation of
    nearest neighbours for the targets)
*/
class MatchPositions {
  static const int num_players_MAX= NUM_PLAYERS;
  static const int num_targets_MAX= 6;

  int best_sol[num_players_MAX];

  struct Dist_2_Player {
    Dist_2_Player() { player= -1; dist= 4*FIELD_BORDER_X; }
    Dist_2_Player(int p,Value d) { player= p; dist= d; }
    int player;
    Value dist;
  };
  Dist_2_Player nearest_from_target[num_targets_MAX][3];
 public:
  MatchPositions() {}
  /** if num_players < num_targets  the will also a solution, but targets
      with lower index will be taken first.
  */

  void match(const Vector* players, int num_players,
	     const Vector* targets, int num_targets);
  /** get the index idx of target, so that players[num] is matched to
      target[idx]. 
      If idx is < 0, the players[num] has no target to go!
   */
  int get_match_for(int num) { return best_sol[num]; }
};

/*****************************************************************************/

/** We need this class for our matching algorithm */
class PlayerMatch{
 public:
  void init(int index, 
	    const Vector me,
	    const Vector* targets, int num_targets);
  double *distances;
  int mapped_to;
  int my_index;
  int shortest_index;
};

/** The class MatchPositions implements another matching algorithm.
    It matchs 10 Players to 10 Positions. First the player with the longest way 
    to go to his closest Position gets assigned that position. We repeat this process
    with the remaining Players and Positions until everyone has a position.
    Note that this is a suboptimal heuristic, but in many cases it works quite 
    well, because everybody will have to run a bit and not, like in many other 
    matching algorithms, one player has to run a very long way while the rest does'nt
    have to move. Additionally the algorithm is quite fast...
*/
class MatchPositionsAlternative{
 private:
  int n_players, n_targets;
  int *map;
  PlayerMatch *player;
  int min_element(double *array, int array_length);
  int max_element(PlayerMatch *array, int array_length);
  int temp_max;
 public:
  /** constructor.
      Please be sure to have at least as many positions as you have players !
  */
  MatchPositionsAlternative(int num_players=10, int num_targets=10);
  /** we try to match num_players players to num_targets targets */
  void match(const Vector* players, int num_players,
	     const Vector* targets, int num_targets);
  /** get the index idx of target, so that players[num] is matched to
      target[idx]. 
      If idx is < 0, the players[num] has no target to go!
   */
  int get_match_for(int num) { return map[num]; }
};

/*****************************************************************************/

/** The third Matching Algorithm does not optimize a certain property for all
    vertices and edges, instead it tries to optimize this property in a way
    that its distribution across the vertices is optimized.

    Example for betting understanding of the above gibberish:
    property=distance, thus not the matching for minimized sum over all
    edges is searched for, but the matching for minimizing the maximum
    distance one(!) player has to walk to its assigned position.

    The class uses the QuickSort Algorithm (by Artur Merke).
*/

class MatchDistribution{

 protected:

  struct val2ord{
    float value;
    int player;
    int position;
  };
  int *map;
  val2ord *v;

  void exchange(int i,int j, val2ord * v);
  void qsort(int size,val2ord * v);
  void convert(float * values, int num_players, int num_positions);

 public:
  void match(float * values, int num_players, int num_positions);

  int get_match_for(int num){return map[num];}

};


/***************************************************************************/
class Policy_Tools {
 private:
  static Value goalshot_goalie_bonusstep;  // goalie is 1 step closer to ball
  static Value goalshot_corner_tolerance; 
  static Value goalshot_consider_age; 
  static Value goalshot_worstcase_consider_age; 
  static Value extremeshot_goalie_maxspeed_percentage; 
  static Value extremeshot_goalie_playersize_percentage; 
  static bool go2ball_list_analytical;
  static Go2Ball_Steps go2ball_list[22];

  static float defense_area;

 public:
  static void init_params();

  /* computes the number of steps to intercept the ball for every player on the pitch and stores the results
     in an array of Go2Ball_Steps structs. set will be sorted by the number of steps to intercept the ball in 
     ascending order.
     Use this set to make a go2ball decision
  */
  static void go2ball_steps_update();
  static Go2Ball_Steps* go2ball_steps_list();
  
  /*  computes the number of steps to intercept a pass to target <target> with velocity <ball_vel> for every player on 
      the pitch and stores the results in an array of Go2Ball_Steps structs. set will be sorted by the number 
      of steps to intercept the ball in ascending order.
      Use this set to select pass candidates
  */
  static void pass_steps_list(Go2Ball_Steps* ball_steps, int player, Vector target, float ball_vel); 

  static float pass_quality(int pass_player, Pass pass);

  static float pass_safety_steps(int pass_player, Vector target, float vel);

  /** computes the 'level of success' of a goal shot with initial ballpos and -vel
      if successful, returns the number of cycles until the fastest opponent gets the ball
      the higher the return, the safer the shot.
      A return of -1 means that an opponent gets the ball before the line
  */
  static float goalshot_success_level(Vector ballpos, Vector target, float vel, 
				      bool worst_case = false);

  /** tests a number of possible goal shots and returns the maximum level
      of success if goalshot is possible, and -1 otherwise 
      In case of success, the initial ballvelocity and direction of the ball
      are also provided.
  */
  static float get_goalshot(float & res_vel, Vector & res_target,float & res_dir,
			    const Vector & ball_pos);
  

  /** computes the earlriest time and position at which the player can
      intercept the ball, the default value player_vel_max= -1.0
      means, that a default value for player velocity will be assumed.  */
  static void intercept_min_time_and_pos(int & res_time, Vector & res_pos,
					 const Vector & ball_pos, 
					 const Vector & ball_vel, 
					 const Vector & player_pos,
					 Value player_vel_max= -1.0, 
					 Value player_size= -1000.0);

  static void intercept_min_time_and_pos(int & res_time, Vector & res_pos,
				const Vector & ball_pos, 
				const Value ballspeed, 
				const Value balldir, 
				const Vector & player_pos, 
				Value player_vel_max= -1.0,
				Value player_size=-1000.0);
  /** computes the earlriest time and position at which 
      player A, dashing with speed in direction dir can be
      intercepted by player B */
  static void intercept_player(int & res_time, Vector & res_pos,
					 const Vector A_pos, 
					 const Value A_speed, 
					 const Value A_dir, 
					 const Vector B_pos,
					 Value player_vel_max= -1.0, 
					 Value player_size= -1000.0);
  
  static void my_team_info(char *seeing, char *knowing);
  static void op_team_info(char *seeing, char *knowing);

  /** compute position where the ball is caught if shot with speed in direction dir
      returns true, if intercept position is inside pitch
      returns if myteam gets the ball, the intercept position and the player number 
  */
  static bool earliest_intercept_pos(const Vector ballpos,const Value speed, const Value dir,
				     Vector &intercept_pos,bool &myteam, int &number);

  /** compute position where the ball intercepted wrt. to the actual game state by calling 
      earliest_intercept_pos(...)
  */
  static Vector Policy_Tools::next_intercept_pos();

  /** returns the teammate player number which is fastest to the ball, when the ball is shot
      from position ballpos with speed and dir */
  static int fastest_player_to_ball(const Vector ballpos,const Value speed, const Value dir,
				    Vector &intercept_pos);
  /** compute the best direction for clearing out of testdir list.
      returns true if either teammate has a chance of getting the ball or ballgain > min_gain (20m)
  */
  static bool get_best_clearance_kick(const Vector ballpos, const int testdirs,
				      const Value testdir[], Value testspeed[], 
				      Value &speed, Value &dir);
  /* defense stuff */
  static void get_opponent_attackers(int *attacker_number, int max_attackers, int op_ball_holder=0);
};


#endif
